import matplotlib.pyplot as plt
import os
import string
import warnings
import numpy as np
from matplotlib import ticker, colors
import glob
import xarray as xr
import tqdm
import glob
from joblib import Parallel, delayed

def plot_point(ax, point_ds, var, ylim = [-0.2, 1], t_idx = None, title = True, xtick = True):
    cmap = plt.cm.get_cmap('tab10')
    handles = []
    if "sensor" in point_ds.data_vars:
        for i, sensor_name in enumerate(np.unique(point_ds["sensor"].values)):
            X = point_ds.time[point_ds.sensor == sensor_name]
            Y = point_ds[var][point_ds.sensor == sensor_name]
            if Y.count("time").values > 0:
                obj = ax.scatter(
                        X, Y,
                        color = cmap(i),
                        label = point_ds.sensor.attrs[str(int(sensor_name))],
                        )
                handles.append(obj)
    else:
        X = point_ds.time
        Y = point_ds[var]
        i = 0 
        if Y.count("time").values > 0:
            obj = ax.scatter(X, Y, color = cmap(i), label = "measurements")
            handles.append(obj)
    obj = ax.plot(point_ds["time"], point_ds[f"{var}_smoothed"], label = "smoothed", color = cmap(i + 1), marker = "X")
    if not isinstance(ylim, type(None)):
        ax.set_ylim(ylim)
    if not isinstance(t_idx, type(None)):
        ax.plot([point_ds["time"][t_idx].values] * 2, ylim, color = "black", linewidth = 5, alpha = 0.3)
    handles.append(obj[0])
    ax.grid()
    ax.set_ylabel(f"{var}")
    ax.set_facecolor("lightgray")
    if not xtick:
        ax.set_xticklabels([])
        ax.set_xlabel("")
    else:
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            ax.set_xticklabels(ax.get_xticklabels(), rotation = 45, ha="right")
        ax.set_xlabel("time [date]")
    if title:
        title_parts = []
        if "pname" in point_ds.attrs.keys():
            name = getattr(point_ds, "pname", "")
            title_parts.append(r"$\bf{" + name + r"}$")
        if "lmbda_sel" in point_ds.data_vars:
            lmbda = point_ds['lmbda_sel']
            title_parts.append(f"Lambda: {lmbda.values:.0E}")
        if "cves" in point_ds.data_vars and "lmbda_sel" in point_ds.data_vars:
            if "lmbda" in point_ds["cves"].dims:
                cve = point_ds['cves'].sel(lmbda = point_ds['lmbda_sel'], method = "nearest").values
            else:
                cve = point_ds["cves"].values
            title_parts.append(f"CVE: {cve:.4f}")
        if "a" in point_ds[f"{var}_smoothed"].attrs.keys():
            title_parts.append(f"a: {point_ds[f'{var}_smoothed'].attrs['a']}")
        if "x" in point_ds.coords:
            lat = float(point_ds.y.values)
            title_parts.append(f"Lat.: {lat:.3f}")
        if "y" in point_ds.coords:
            lon = float(point_ds.x.values)
            title_parts.append(f"Lon.: {lon:.3f}")
        full_title = ", ".join(title_parts)
        ax.set_title(full_title)
    ax.legend(handles = handles, ncols = 4, loc = "lower center")
    return handles

def plot_map(ax, da, points = None, cmap = "RdBu_r", ylim = [-1.0, 1.0], ytick = True, xtick = True, norm = None):
    assert da.ndim == 2
    im = ax.pcolormesh(da.x, da.y, da, cmap = cmap, vmin = ylim[0], vmax = ylim[1], norm=norm)
    ax.grid()
    ax.set_facecolor("lightgray")
    ax.set_title(da.name)
    if not isinstance(points, type(None)):
        ax.scatter(points[0], points[1], marker = "o", s = 100, edgecolors="black", linewidths = 2, c = [(1,1,1,0)], zorder = 100)
        for x,y,name in zip(*points):
            ax.annotate(name, (x,y), (30,0), textcoords = 'offset pixels', fontsize = 15)
    if not ytick:
        ax.set_yticklabels([])
    else:
        ax.set_ylabel("Lat. [DD]")
        ax.yaxis.set_major_formatter(ticker.StrMethodFormatter("{x:.3f}"))
        ax.yaxis.set_major_locator(plt.MaxNLocator(4))
    if not xtick:
        ax.set_xticklabels([])
    else:
        ax.set_xlabel("Lon. [DD]")
        ax.xaxis.set_major_formatter(ticker.StrMethodFormatter("{x:.3f}"))
        ax.xaxis.set_major_locator(plt.MaxNLocator(4))

    return im

def plot_video_frame(ds, points, var, t_idx, folder):

    if not os.path.exists(folder):
        os.makedirs(folder)

    panels = [["A", "B"]]
    for i in range(len(points[0])):
        panels.append([str(i), str(i)])

    fig, axes = plt.subplot_mosaic(panels, figsize = (10, (len(points[0]) + 1)*4), dpi = 16**2)

    im1 = plot_map(axes["A"], ds["ndvi"].isel(time=t_idx), points=points)
    im2 = plot_map(axes["B"], ds["ndvi_smoothed"].isel(time=t_idx), ytick = False, points = points)

    for i, (lon, lat, name) in enumerate(zip(*points)):
        xtick = False if i+1 < len(points[0]) else True
        point_ds = ds.sel({"x": lon, "y": lat}, method = "nearest")
        point_ds = point_ds.assign_attrs({"pname": name})
        _ = plot_point(axes[str(i)], point_ds, var, t_idx = t_idx, title = True, xtick = xtick)

    fig.colorbar(im1)
    fig.colorbar(im2)
    fig.subplots_adjust(left=None, bottom=None, right=None, top=0.91, wspace=None, hspace=0.3)
    fig.suptitle(np.datetime_as_string(ds.time[t_idx], unit='m'))
    fig.savefig(os.path.join(folder, f"{t_idx:>06}.png"))
    plt.close(fig)

def make_overview(ds, var, plot_folder, points = None, point_method = "equally_spaced", n = 3, offset = 0.1, **kwargs):

    if isinstance(points, type(None)):
        if point_method == "equally_spaced":
            points = create_points(ds, n = n, offset = offset)
        elif point_method == "worst":
            points = create_worst_points(ds, f"{var}_smoothed", "time", n = n)

    if not os.path.exists(plot_folder):
        os.makedirs(plot_folder)

    npanels = np.sum(np.isin(["lmbda_sel", var, "cves"], ds.data_vars))
    panels = [list(string.ascii_uppercase[:npanels])]
    for i in range(len(points[0])):
        panels.append([str(i)] * len(panels[0]))

    fig, axes = plt.subplot_mosaic(panels, figsize = (12, (len(points[0]) + 1)*4), dpi = 16**2)

    if var in ds.data_vars:
        ds["counts"] = ds[var].count(dim = "time")
        im1 = plot_map(axes[panels[0].pop(0)], ds["counts"], ylim = [None, None], cmap = "viridis", points = points)
        fig.colorbar(im1, label = "No. measurements [-]")

    if np.all(np.isin(["cves", "lmbda_sel"], ds.data_vars)):
        if "lmbda" in ds["cves"].dims:
            ds["cve"] = ds["cves"].sel(lmbda = ds["lmbda_sel"], method = "nearest")
        else:
            ds["cve"] = ds["cves"]
        im2 = plot_map(axes[panels[0].pop(0)], ds["cve"], ylim = [0, np.nanpercentile(ds["cve"], 95)], cmap = "viridis", ytick = False, points = points)
        fig.colorbar(im2, label = "Cross-val. Standard Error [-]")
    
    if "lmbda_sel" in ds.data_vars:
        if ds["lmbda_sel"].ndim == 0:
            ds["lmbda_sel"] = xr.ones_like(ds["counts"]) * ds["lmbda_sel"]
        im3 = plot_map(axes[panels[0].pop(0)], ds["lmbda_sel"], ylim = [None, None], cmap = "viridis", ytick = False, norm = colors.LogNorm(), points = points)
        fig.colorbar(im3, label = "Lambda [-]")

    for i, (lon, lat, name) in enumerate(zip(*points)):
        xtick = False if i+1 < len(points[0]) else True
        point_ds = ds.sel({"x": lon, "y": lat}, method = "nearest")
        point_ds = point_ds.assign_attrs({"pname": name})
        _ = plot_point(axes[str(i)], point_ds, var, ylim = None, t_idx = None, title = True, xtick = xtick)

    fig.subplots_adjust(left=None, bottom=None, right=None, top=0.95, wspace=None, hspace=0.3)
    fig.savefig(os.path.join(plot_folder, f"{var}_overview.png"))
    plt.close(fig)

def make_video(ds, points, fh, var, xdim = "time", new_x = None, parallel = True):

    folder = os.path.split(fh)[0]

    if not os.path.exists(folder):
        os.makedirs(folder)

    idxs = np.arange(0, ds[xdim].size, 1)
    if not isinstance(new_x, type(None)):
        idxs = idxs[~np.isin(ds[xdim], new_x)]

    if not parallel:
        for i in tqdm.tqdm(idxs):
            plot_video_frame(ds, points, var, i, folder)
    else:
        _ = Parallel(n_jobs=4)(delayed(plot_video_frame)(ds, points, var, i, folder) for i in idxs)
    
    files = np.sort(glob.glob(os.path.join(folder, "[0-9]" * 6 + ".png")))
    
    create_video(files, fh)

def create_points(ds, n = 3, offset = 0.1):
    
    if isinstance(offset, float):
        offsetx = np.floor(ds.x.size * offset).astype(int)
        offsety = np.floor(ds.y.size * offset).astype(int)
    elif isinstance(offset, int):
        offsetx = offset.copy()
        offsety = offset.copy()

    offsetx = np.min([offsetx, np.floor((ds.x.size - n) / 2).astype(int)])
    offsety = np.min([offsety, np.floor((ds.y.size - n) / 2).astype(int)])

    lons_idxs = np.linspace(0+offsetx, ds.x.size-(1+offsetx), n, dtype=int)
    lats_idxs = np.linspace(0+offsety, ds.y.size-(1+offsety), n, dtype=int)

    lons = ds.x.isel(x=lons_idxs).values
    lats = ds.y.isel(y=lats_idxs).values
    
    ys, xs = np.meshgrid(lats, lons)
    
    ys = ys.flatten().tolist()
    xs = xs.flatten().tolist()
    names = [f"P{i:>02}" for i in range(1, len(xs)+1)]
    return (xs, ys, names)

def create_worst_points(ds, var, dim, n = 8):

    ds = np.abs(ds).max(dim = dim)

    stacked_ds = ds.stack({"point": ds[var].dims})[var]

    selected = stacked_ds.isel(point = stacked_ds.argsort().values).isel(point = slice(-n-1, -1))

    points = (selected.x.values, selected.y.values, [f"P{i:>02}" for i in range(1, n+1)])
    return points

def create_video(files, video_fh, fps = 4, remove_files = True):

    try:
        import imageio.v2 as imageio
    except ImportError:
        print("--> Install `imageio` to automatically create a video")
        return

    try:
        with imageio.get_writer(video_fh, fps = fps) as writer:
            for im in files:
                writer.append_data(imageio.imread(im))
        if remove_files:
            for fn in files:
                try:
                    os.remove(fn)
                except PermissionError:
                    continue
    except ValueError as e:
        msg = getattr(e, "args", ["None"])
        if "Based on the extension, the following" in msg[0]:
            fn, ext = os.path.splitext(video_fh)
            if ext == ".mp4":
                print("--> Creating `.gif` file, install `imageio_ffmpeg` to create `.mp4`.")
                create_video(files, fn + ".gif", fps = fps)
            else:
                print("--> Unable to create video with requested extension.")
                print(e)
        else:
            raise e



"""
https://nextcloud.lsasvcs.ipma.pt/s/QSABgnG4dZGBo5W?dir=undefined&path=%2FPUM-Product_User_Manual&openfile=27089
https://nextcloud.lsasvcs.ipma.pt/s/bA9gYoa5mQX2yJw?dir=undefined&path=%2FPUM-Product_User_Manual&openfile=27251
https://lsa-saf.eumetsat.int/en/data/products/radiation/
"""

import pywapor
import pywapor.collect.protocol.crawler
import pywapor.collect.protocol.opendap
from pywapor.enhancers.apply_enhancers import apply_enhancers
from pywapor.general.processing_functions import open_ds, save_ds, remove_ds, is_corrupt_or_empty
from pywapor.general.logger import log
from pywapor.collect import accounts
from pywapor.collect.protocol.projections import get_crss
from joblib import Parallel, delayed
import multiprocessing
from joblib import Memory
import xarray as xr
import tqdm
import re
import os
import datetime
import copy
import posixpath
import requests
import pandas as pd
import numpy as np
from pywapor.general import bitmasks

# NOTE https://stackoverflow.com/questions/75202475/joblib-persistence-across-sessions-machines
def cache(mem, module, **mem_kwargs):
    def cache_(f):
        f.__module__ = module
        f.__qualname__ = f.__name__
        return mem.cache(f, **mem_kwargs)
    return cache_

def apply_qa(ds, var, masker = "qa", product_name = "MSG_MDSSFTD", flags = ["clear_ok", "cloudy_sky_method_ok"]):
    flag_bits = bitmasks.LSASAF_qa_translator(product_name)
    mask = bitmasks.get_mask(ds[masker].astype("uint8"), flags, flag_bits)
    new_data = ds[var].where(mask, np.nan)
    ds[var] = new_data
    return ds

def night_is_zero(ds, var, masker = "qa", product_name = "MSG_MDSSFTD"):
    flag_bits = bitmasks.LSASAF_qa_translator(product_name)
    mask_day_night = bitmasks.get_mask(ds[masker].astype("uint8"), ["clear_night"], flag_bits)
    mask_land_cont_water = bitmasks.get_mask(ds[masker].astype("uint8"), ["land", "continental_water"], flag_bits)
    mask = mask_land_cont_water & mask_day_night
    new_data = ds[var].where(~mask, 0)
    ds[var] = new_data
    return ds

def jouleperday_to_watt(ds, var):
    ds[var] = ds[var] / 86400
    return ds

def celsius_to_kelvin(ds, var):
    ds[var] = ds[var] + 273.15
    return ds

def fn_func(product_name, tile):
    fn = os.path.split(tile)[-1]
    return fn

def url_func(product_name, tile):
    return tile

def default_vars(product_name, req_vars):

    variables = {

        "MSG_MDSSFTD": { # MSG Total and Diffuse DSSF, https://nextcloud.lsasvcs.ipma.pt/s/bA9gYoa5mQX2yJw
                    "DSSF_TOT": [("time", "lat", "lon"), "ra_flat"],
                    "FRACTION_DIFFUSE": [("time", "lat", "lon"), "diffuse_fraction"],
                    "quality_flag": [("time", "lat", "lon"), "qa"],
                        },

        "MSG_MDIDSSF": { # MSG Daily DSSF, https://nextcloud.lsasvcs.ipma.pt/s/QSABgnG4dZGBo5W
                    "DSSF": [("time", "lat", "lon"), "ra_flat"],
                    "max_nslots_missing": [("time", "lat", "lon"), "max_nslots_missing"],
                    "missing_values_percent": [("time", "lat", "lon"), "missing_values_percent"],
                    "weight_missing_values_percent": [("time", "lat", "lon"), "weight_missing_values_percent"],
        },

        #########

        "MSG_MLST": {
                    "LST": [("time", "lat", "lon"), "lst"], # inst. lst, Celsius
                    "quality_flag": [("time", "lat", "lon"), "qa"],
                    "standard_error": [("time", "lat", "lon"), "error"],
        },

        "MSG_MLST-AS": {
                    "MLST-AS": [("time", "lat", "lon"), "lst"], # inst. lst (all-sky, i.e. clear AND cloudy), Celsius
                    "quality_flag": [("time", "lat", "lon"), "qa"],
                    "CMa": [("time", "lat", "lon"), "cloud_mask"],
        },

        "MSG_MLST-ASv2": {
                    "MLST-AS": [("time", "lat", "lon"), "lst"], # inst. lst (all-sky, i.e. clear AND cloudy), Celsius
                    "quality_flag": [("time", "lat", "lon"), "qa"],
                    "CMa": [("time", "lat", "lon"), "cloud_mask"],
        },

        "MSG_METREF": {
                    "METREF": [("time", "lat", "lon"), "et_ref_24_mm"], # mm/day
                    "quality_flag": [("time", "lat", "lon"), "qa"],
        },

        "MSG_MH": {
                    "MH": [("time", "lat", "lon"), "h_i"], # sensible heat flux, W/m2
                    "quality_flag": [("time", "lat", "lon"), "qa"],
        },

        "MSG_MLE": {
                    "MLE": [("time", "lat", "lon"), "lh_i"], # latent heat flux, W/m2
                    "quality_flag": [("time", "lat", "lon"), "qa"],
        },

    }

    req_dl_vars = {

        'MSG_MDSSFTD': {
                    'ra_flat': ['DSSF_TOT'],
                    'diffuse_fraction': ['FRACTION_DIFFUSE'],
                    'qa': ['quality_flag']
                    },

        'MSG_MDIDSSF': {
                'ra_flat': ['DSSF'],
                'max_nslots_missing': ['max_nslots_missing'],
                'missing_values_percent': ['missing_values_percent'],
                'weight_missing_values_percent': ['weight_missing_values_percent']
        },

        'MSG_DLST-MAX10D': {
                'lst': ['LST_MAX'],
                'nvalids': ['NUM_VALID'],
                'qa': ['quality_flag'],
                'error': ['standard_error']
        },

        'MSG_MLST': {
                'lst': ['LST'],
                'qa': ['quality_flag'],
                'error': ['standard_error']
        },

        'MSG_MLST-AS': {
                'lst': ['MLST-AS'],
                'qa': ['quality_flag'],
                'cloud_mask': ['CMa']
        },

        'MSG_MLST-ASv2': {
                'lst': ['MLST-AS'],
                'qa': ['quality_flag'],
                'cloud_mask': ['CMa']
        },

        'MSG_METREF': {
                'et_ref_24_mm': ['METREF'], 
                'qa': ['quality_flag']
        },

        'MSG_MH': {
                'h_i': ['MH'], 
                'qa': ['quality_flag']
        },

        'MSG_MLE': {
                'lh_i': ['MLE'], 
                'qa': ['quality_flag']
        },

    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}

    return out

def default_post_processors(product_name, req_vars):

    post_processors = {
        'MSG_MDSSFTD': {'ra_flat': [], 'diffuse_fraction': [], 'qa': []},
        'MSG_MDIDSSF': {'ra_flat': [jouleperday_to_watt], 'max_nslots_missing': [], 'missing_values_percent': [], 'weight_missing_values_percent': []},
        'MSG_MLST': {'lst': [celsius_to_kelvin], 'qa': [], 'error': []},
        'MSG_MLST-AS': {'lst': [celsius_to_kelvin], 'qa': [], 'cloud_mask': []},
        'MSG_MLST-ASv2': {'lst': [celsius_to_kelvin], 'qa': [], 'cloud_mask': []},
        'MSG_METREF': {'et_ref_24_mm': [], 'qa': []},
        'MSG_MH': {'h_i': [], 'qa': []},
        'MSG_MLE': {'lh_i': [], 'qa': []},
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def download(folder, latlim, lonlim, timelim, product_name, req_vars,
                 variables = None, post_processors = None):

    folder = os.path.join(folder, "LSASAF")

    satellite, product_name_ = product_name.split("_")
    x = product_name_.split("-")
    if len(x) == 1:
        product_name__ = x[0]
        extra_filter = None
    elif len(x) == 2:
        product_name__ = x[0]
        extra_filter = x[1]
    else:
        raise ValueError
    format = "NETCDF"

    if not os.path.exists(folder):
        os.makedirs(folder)

    # NOTE paths on windows have a max length, this extends the max length, see
    # here for more info https://learn.microsoft.com/en-us/windows/win32/fileio/maximum-file-path-limitation?tabs=registry
    if os.name == "nt": 
        cachedir = "\\\\?\\" + os.path.join(os.path.abspath(folder), "cache")
    else:
        cachedir = os.path.join(folder, "cache")

    memory = Memory(cachedir, verbose=0)

    fp = os.path.join(folder, f"{product_name}.nc")
    req_vars_orig = copy.deepcopy(req_vars)
    if os.path.isfile(fp):
        existing_ds = open_ds(fp)
        req_vars_new = list(set(req_vars).difference(set(existing_ds.data_vars)))
        if len(req_vars_new) > 0:
            req_vars = req_vars_new
            existing_ds = existing_ds.close()
        else:
            return existing_ds[req_vars_orig]
        
    spatial_buffer = True
    if spatial_buffer:
        res = {"MSG": 0.05}.get(satellite, "define_the_sat_res_please")
        latlim = [latlim[0] - res, latlim[1] + res]
        lonlim = [lonlim[0] - res, lonlim[1] + res]

    coords = {"x": ["lon", lonlim], "y": ["lat", latlim[::-1]], "t": ["time", timelim]}

    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    data_source_crs = get_crss("WGS84")

    timedelta = {
        "MSG_MDIDSSF": np.timedelta64(12, "h"),
        "MSG_METREF": np.timedelta64(12, "h"),
     }.get(product_name, None)
    
    un, pw = accounts.get("LSASAF")
    session = requests.Session()
    session.auth = (un, pw)

    @cache(memory, "LSASAF")
    def scrape(base_url):
        catalog_url = posixpath.join(base_url, "catalog.html")
        resp = session.get(catalog_url)
        resp.raise_for_status()
        return resp.content
    
    regex = r'nc"><code>(.*?)<\/code>'
    dates = pd.date_range(timelim[0], timelim[1], freq = "D")
    urls = list()
    for date in list(dates[:-1]):
        base_url = f"https://thredds.lsasvcs.ipma.pt/thredds/catalog/{satellite}/{product_name__}/{format}/{date.year}/{date.month:>02}/{date.day:>02}"
        content = scrape(base_url)
        fns = re.findall(regex, str(content))
        if not isinstance(extra_filter, type(None)):
            fns = [x for x in fns if f"-{extra_filter}_" in x]
        urls_ = [posixpath.join(base_url, fn).replace("/catalog/", "/dodsC/") for fn in fns]
        urls += urls_

    log.info(f"--> Found {len(urls)} relevant tiles.")

    urls_ = [url for url in urls if not os.path.isfile(os.path.join(folder, os.path.split(url)[-1]))]
    urls__ = [url for url in urls if url not in urls_]

    log.info(f"--> Downloading remaining {len(urls_)} tiles.").add()

    def download_tile(url, folder):
        fn = os.path.split(url)[-1]
        fp_ = os.path.join(folder, fn)
        corrupt = False
        if os.path.isfile(fp_):
            corrupt = is_corrupt_or_empty(fp_)
            if corrupt:
                log.info(f"--> Removing corrupt or empty file `{os.path.split(fp_)[-1]}`.")
                remove_ds(fp_)
            else:
                log.info(f"--> Opening existing file `{os.path.split(fp_)[-1]}`.")
                ds_ = open_ds(fp_)
        if corrupt or not os.path.isfile(fp_):
            url_ = url.replace("https://", f"https://{un}:{pw}@")
            ds_ = pywapor.collect.protocol.opendap.download_xarray(url_, fp_, coords, variables, {}, 
                                data_source_crs = data_source_crs, timedelta=timedelta, parallel=False)
        log.sub()
        return ds_

    n_jobs = max(1, multiprocessing.cpu_count() - 1)
    parallel = True
    if parallel:
        dss = Parallel(n_jobs=n_jobs)(delayed(download_tile)(url, folder) for url in tqdm.tqdm(urls_, leave = False))
    else:
        dss = list()
        for i, url in enumerate(urls_):
            fn = os.path.split(url)[-1]
            log.info(f"--> ({i+1}/{len(urls)}) Collecting `{fn}`.").add()
            dss.append(download_tile(url, folder))
            log.sub()
    
    for url in urls__:
        fp_ = os.path.join(folder, os.path.split(url)[-1])
        dss.append(open_ds(fp_))

    log.sub()

    # Combine all data.
    ds = xr.concat(dss, dim = "time").sortby("time")

    # Apply product specific functions.
    ds = apply_enhancers(post_processors, ds)

    # Save final output.
    out_ = save_ds(ds, fp, encoding = "initiate", label = "Saving netCDF.")

    for x in dss:
        remove_ds(x)

    return out_[req_vars_orig]

if __name__ == "__main__":

    from pywapor.general.logger import adjust_logger
    import matplotlib.pyplot as plt

    tests = [
        ('MSG_MDSSFTD', ['ra_flat', 'diffuse_fraction', 'qa']),
        ('MSG_MDIDSSF', ['ra_flat', 'max_nslots_missing', 'missing_values_percent', 'weight_missing_values_percent']),
        ('MSG_MLST', ['lst', 'qa', 'error']),
        ('MSG_MLST-AS', ['lst', 'qa', 'cloud_mask']),
        ('MSG_MLST-ASv2', ['lst', 'qa', 'cloud_mask']),
        ('MSG_METREF', ['et_ref_24_mm', 'qa']),
        ('MSG_MH', ['h_i', 'qa']),
        ('MSG_MLE', ['lh_i', 'qa']),
    ]

    folder = r"/Users/hmcoerver/Local/ra_test"

    adjust_logger(True, folder, "INFO")

    latlim = [30.168923, 36.520673]
    lonlim = [29.787384, 44.625092]
    timelim = [datetime.date(2023,2,5), datetime.date(2023,2,6)]
    # timelim = [datetime.date(2023,2,1), datetime.date(2023,2,13)]

    variables = None
    post_processors = None

    dss = list()
    for product_name, req_vars in tests:
        log.info(f"{product_name}, {req_vars}")
        ds_x = download(folder, latlim, lonlim, timelim, product_name, req_vars)
        dss.append(ds_x)

    # product_name = "MSG_MDIDSSF"
    # req_vars = ["ra_flat", "max_nslots_missing", "missing_values_percent", "weight_missing_values_percent"]
    # ds1 = download(folder, latlim, lonlim, timelim, product_name, req_vars)

    # product_name = "MSG_MDSSFTD"
    # req_vars = ["ra_flat", "qa"]
    # ds2 = download(folder, latlim, lonlim, timelim, product_name, req_vars)

    # ds3 = pywapor.collect.product.ERA5.download(
    #                                     folder, 
    #                                     latlim, 
    #                                     lonlim, 
    #                                     timelim, 
    #                                     product_name="sis-agrometeorological-indicators", 
    #                                     req_vars=["ra_flat"]
    #                             )

    # ds4 = pywapor.collect.product.MERRA2.download(
    #                                     folder, 
    #                                     latlim, 
    #                                     lonlim, 
    #                                     timelim, 
    #                                     product_name="M2T1NXRAD.5.12.4", 
    #                                     req_vars=["ra_flat"]
    #                             )

    # ds5 = pywapor.collect.product.GEOS5.download(
    #                                     folder, 
    #                                     latlim, 
    #                                     lonlim, 
    #                                     timelim, 
    #                                     product_name="tavg1_2d_rad_Nx", 
    #                                     req_vars=["ra_flat"]
    #                             )
    
    # dss = [ds1, ds2, ds3, ds4, ds5]

    # fig, axs = plt.subplots(len(dss), 1, figsize = (6, 15), sharex=True, sharey=True, dpi = 300)

    # var = "ra_flat"

    # timeslice = [np.datetime64("2023-02-01 00:00:00"), np.datetime64("2023-02-03 00:00:00")]

    # for ax, ds in zip(axs, dss):
    #     source_name = os.path.split(os.path.split(ds.encoding["source"])[0])[-1]
    #     prod_name = os.path.splitext(os.path.split(ds.encoding["source"])[-1])[0]
    #     ds_ = ds[var].sel(time = slice(*timeslice))
    #     n = ds_.sizes["time"]
    #     ds__ = ds_.mean(dim = "time")
    #     ds__.attrs["units"] = "W m-2"
    #     ds__.plot(ax = ax, vmax = 220)
    #     ax.set_title(f"{source_name}.{prod_name}, n = {n}")
    #     xlabel = ax.get_xlabel()
    #     ax.set_xlabel("")
    #     ax.set_facecolor("lightgray")

    # plt.xlabel(xlabel)
    # plt.suptitle(f"{var}\n[{timeslice[0]} - {timeslice[1]}]")

    # plt.savefig(os.path.join(folder, "ra_flat_comparison.png"))



"""
    The roughness module contains all functions related to surface roughness.

"""
import numpy as np
import xarray as xr

def roughness_length(lai, z_oro, z_obst, z_obst_max, land_mask=1):
    r"""
    Computes the surface roughness length. The roughness length is related to
    the roughness characteristics. For the logarithmic wind-profile the
    surface roughness length is the height at which the wind speed is zero.
    The roughness length is calculated differently for different types of land use.

    Land use is classified as follows:

    0. no data
    1. land
    2. water
    3. urban

    .. math ::

        z_{0,m}=\begin{cases}
        \begin{array}{cc}
        0 & l=0\\
        z_{0,m} & l=1\\
        0.0001 & l=2\\
        \frac{1}{7} \cdot z_{obst,max}+z_{oro} & l=3
        \end{array}\end{cases}


    Parameters
    ----------
    lai : float
        leaf area index, 
        :math:`I_{lai}`
        [-]
    z_oro : float
        orographic roughness, 
        :math:`z_{oro}`
        [m]
    z_obst : float
        obstacle height, 
        :math:`z_{obst}`
        [m]
    z_obst_max : float
        maximum obstacle height, 
        :math:`z_{obst,max}`
        [m]
    land_mask : int
        land use classification, 
        :math:`l`
        [-]

    Returns
    -------
    z0m : float
        roughness length, 
        :math:`z_{0,m}`
        [m]

    Examples
    --------
    >>> import ETLook.roughness as roughness
    >>> roughness.roughness_length(0.4)
    0.34179999999999999

    """

    def veg_roughness(z_obst, disp, z_obst_max, lai, z_oro):
        veg = 0.193
        z_dif = z_obst-disp

        term1 = np.minimum(0.41**2/((np.log(z_dif/(0.002*z_obst_max))+veg)**2), 1.0) + 0.35*lai/2
        term2 = np.exp(0.41/np.minimum(np.sqrt(term1), 0.3)-veg)

        return z_dif/term2+z_oro

    # roughness length specific displacement height
    disp = z_obst * (1 - (1 - np.exp(-np.sqrt(12 * lai))) / (np.sqrt(12 * lai)))
    if isinstance(disp, xr.DataArray):
        disp = xr.where(lai == 0, 0.0, disp)
    else:
        disp[lai == 0] = 0

    z0m = (1. / 7.) * z_obst_max + z_oro
    if isinstance(disp, xr.DataArray):
        z0m = xr.where(land_mask == 0, 0.0, z0m)
    else:
        z0m[land_mask == 0] = 0

    if isinstance(disp, xr.DataArray):
        z0m = xr.where(land_mask == 1, veg_roughness(z_obst, disp, z_obst_max, lai, z_oro), z0m)
        z0m = xr.where(land_mask == 2, 0.0001, z0m)
    else:
        z0m = np.where(land_mask == 1, veg_roughness(z_obst, disp, z_obst_max, lai, z_oro), z0m)
        z0m[land_mask == 2] = 0.0001

    return z0m


def obstacle_height(ndvi, z_obst_max, ndvi_obs_min=0.25,
                    ndvi_obs_max=0.75, obs_fr=0.25):
    r"""
    Computes the obstacle height. The ndvi is used to limit the obstacle
    height.

    .. math ::
        z_{obst}	=	\begin{cases}
        \begin{array}{cc}
        f_{obs}z_{obst,max} & I_{ndvi}\leq I_{ndvi,obs,min}\\
        z_{obst,max}\left(f_{obs}+\left(1-f_{obs}\right)\left
        (\frac{I_{ndvi}-I_{ndvi,obs,min}}
        {I_{ndvi,obs,max}-I_{ndvi,obs,min}}\right)\right) &
        I_{ndvi}>I_{ndvi,obs,min}\&I_{ndvi}<I_{ndvi,obs,max}\\
        z_{obst,max} & I_{ndvi}\geq I_{ndvi,obs,max}
        \end{array}\end{cases}

    Parameters
    ----------
    ndvi : float
        normalized difference vegetation index, 
        :math:`I_{ndvi}`
        [-]
    ndvi_obs_min : float
        normalized difference vegetation index @ min obstacle height, 
        :math:`I_{ndvi,obs,min}`
        [-]
    ndvi_obs_max : float
        normalized difference vegetation index @ max obstacle height, 
        :math:`I_{ndvi,obs,max}`
        [-]
    obs_fr : float
        ratio of minimum and maximum obstacle height, 
        :math:`f_{obs}`
        [-]
    z_obst_max : float
        maximum obstacle height
        :math`z_{obst,max}`
        [m]

    Returns
    -------
    z_obst : float
        obstacle height, 
        :math:`z_{obst}`
        [m]

    Examples
    --------
    >>> import ETLook.roughness as roughness
    >>> roughness.obstacle_height(0.4, 2.0)
    0.95

    """

    def frac_func(n):
        return (obs_fr + (1-obs_fr)*(n-ndvi_obs_min)/
                (ndvi_obs_max-ndvi_obs_min))

    if isinstance(ndvi, xr.DataArray):
        obs_height = xr.where(ndvi <= ndvi_obs_min, obs_fr*z_obst_max,  z_obst_max)
        obs_height = xr.where(np.logical_and(ndvi > ndvi_obs_min, ndvi < ndvi_obs_max), frac_func(ndvi)*z_obst_max, obs_height)
        obs_height = xr.where(ndvi >= ndvi_obs_max, z_obst_max, obs_height)
    else:
        obs_height = z_obst_max.copy()
        obs_height = np.where(ndvi <= ndvi_obs_min, obs_fr*z_obst_max,  obs_height)
        obs_height = np.where(np.logical_and(ndvi > ndvi_obs_min, ndvi < ndvi_obs_max), frac_func(ndvi)*z_obst_max, obs_height)
        obs_height = np.where(ndvi >= ndvi_obs_max, z_obst_max, obs_height)

    return obs_height

def displacement_height(lai, z_obst, land_mask=1, c1=1):
    r"""
    Computes the displacement height. The lai is used to limit the displacement
    height. It is defined differently for different types of landuse.

    Land use is classified as follows:

    0. no data
    1. land
    2. water
    3. urban

    .. math ::

        z_{disp}=\begin{cases}
        \begin{array}{cc}
        0 & l=0\\
        z_{obst}\left(1-\frac{1-\exp\left(-\sqrt{c_{1} \cdot I_{lai}}\right)}
        {\sqrt{c_{1} \cdot I_{lai}}}\right) & l=1\\
        0 & l=2\\
        \frac{2}{3} \cdot z_{obst} & l=3
        \end{array}\end{cases}

    Parameters
    ----------
    lai : float
        leaf area index, 
        :math:`I_{lai}`
        [-]
    z_obst : float
        obstacle height, 
        :math:`z_{obst}`
        [m]
    land_mask : int
        land use classification, 
        :math:`l`
        [-]
    c1 : float
        exponential growth rate displacement height function, 
        :math:`c_1`
        [-]


    Returns
    -------
    disp : float
        displacement height, 
        :math:`{disp}`
        [m]

    Examples
    --------
    >>> import ETLook.roughness as roughness
    >>> roughness.displacement_height(0.4, 2.0)
    0.51779495

    """

    def disp_func(l):
        return z_obst * (1-(1-np.exp(-np.sqrt(c1*l)))/np.sqrt(c1*l))

    if isinstance(land_mask, xr.DataArray):
        disp = xr.zeros_like(land_mask)
        disp = xr.where(land_mask == 1, disp_func(lai), disp)
        disp = xr.where(land_mask == 2, 0, disp)
        disp = xr.where(land_mask == 3, (2./3.)*z_obst, disp)
        disp = xr.where(np.isnan(disp), 0.0, disp)
    else:
        disp = np.zeros_like(land_mask)
        disp = np.where(land_mask == 1, disp_func(lai), disp)
        disp = np.where(land_mask == 2, 0, disp)
        disp = np.where(land_mask == 3, (2./3.)*z_obst, disp)
        disp[np.isnan(disp)] = 0
    
    return disp



import rasterio

def get_crss(name = None):
    crss = {"MODIS": rasterio.crs.CRS.from_wkt("""
            PROJCRS["unnamed",
                BASEGEOGCRS["Unknown datum based upon the custom spheroid",
                    DATUM["Not_specified_based_on_custom_spheroid",
                        ELLIPSOID["Custom spheroid",6371007.181,0,
                            LENGTHUNIT["metre",1,
                                ID["EPSG",9001]]]],
                    PRIMEM["Greenwich",0,
                        ANGLEUNIT["degree",0.0174532925199433,
                            ID["EPSG",9122]]]],
                CONVERSION["unnamed",
                    METHOD["Sinusoidal"],
                    PARAMETER["Longitude of natural origin",0,
                        ANGLEUNIT["degree",0.0174532925199433],
                        ID["EPSG",8802]],
                    PARAMETER["False easting",0,
                        LENGTHUNIT["Meter",1],
                        ID["EPSG",8806]],
                    PARAMETER["False northing",0,
                        LENGTHUNIT["Meter",1],
                        ID["EPSG",8807]]],
                CS[Cartesian,2],
                    AXIS["easting",east,
                        ORDER[1],
                        LENGTHUNIT["Meter",1]],
                    AXIS["northing",north,
                        ORDER[2],
                        LENGTHUNIT["Meter",1]]]
            """),

        "WGS84": rasterio.crs.CRS.from_epsg(4326),
    }
    if not isinstance(name, type(None)):
        return crss[name]
    else:
        return crss


import pywapor
import os
import re
import requests
import numpy as np
import glob
import warnings
import json
import importlib
from osgeo import gdal
from functools import partial
from pywapor.general.logger import log, adjust_logger
from pywapor.collect.downloader import collect_sources
from pywapor.general.compositer import time_bins
from pywapor.general.processing_functions import adjust_timelim_dtype, func_from_string, open_ds, remove_ds, is_corrupt_or_empty, has_wrong_bb_or_period

class Configuration():

    variable_categories = {
        "optical": ["ndvi", "r0", "mndwi", "psri", "vari_red_edge", "bsi", "nmdi", "green", "nir"],
        "thermal": ["bt", "lst"],
        "solar radiation": ["ra_flat"],
        "precipitation": ["p"],
        "elevation": ["z", "slope", "aspect"],
        "meteorological": ["t_air", "t_air_min", "t_air_max", "u", "vp", "p_air", "p_air_0", "u2m", "v2m", "qv", "wv", "t_dew"],
        "statics": ["lw_slope", "lw_offset", "z_obst_max", "rs_min", "land_mask", "vpd_slope", "t_opt", "t_amp", "t_amp_year", "rn_slope", "rn_offset", "z_oro"],
        "soil moisture": ["se_root"],
    }

    category_variables = {}
    for x in [{var: cat for var in varis} for cat, varis in variable_categories.items()]:
        category_variables = {**category_variables, **x}

    se_root_vars = [
        [("ndvi",)], 
        [("bt", "lst"), ("lst",), ("bt",)], 
        [("u",), ("u2m", "v2m")],
        [("p_air",)], 
        [("p_air_0",)], 
        [("t_air",)], 
        [("wv",)], 
        [("t_dew",), ("qv",)]
        ]
    
    et_look_vars = [
        [("ndvi",)],
        [("r0",)],

        [("se_root",)], 

        [("p",)],
        [("z", "slope", "aspect"), ("z")],
        [("ra_flat",)],

        [("u",), ("u2m", "v2m")],
        [("p_air",)], 
        [("p_air_0",)], 
        [("t_air",)],
        [("t_air_min",)], 
        [("t_air_max",)], 
        [("t_dew",), ("qv",)],

        [("rn_offset",)],
        [("rn_slope",)],
        [("t_amp",),],
        [("t_opt",)],
        [("vpd_slope",)],
        [("lw_offset",)],
        [("lw_slope",)],

        [("rs_min",), ()],
        [("land_mask",), ()],
        [("z_obst_max",), ()],
        [("z_oro",), ()],
        ]
    
    @staticmethod
    def source_func(x):
        return x if "FILE:" in x else x.split(".")[0]
    
    @staticmethod
    def pname_func(x):
        return "none" if "FILE:" in x else ".".join(x.split(".")[1:])

    def __init__(self, full = None, summary = None, se_root = None, 
                 et_look = None):
        self.full = full
        self.summary = summary
        self.se_root = se_root
        self.et_look = et_look

    def __repr__(self):
        summary = self.summary.copy()

        example_product = summary.pop('_EXAMPLE_')
        whittaker = summary.pop('_WHITTAKER_')
        sharpen = summary.pop('_ENHANCE_')

        base = "\n".join([f"--> {cat.title()} data:\n    > `{'`, `'.join(sources)}`" for cat, sources in summary.items()])

        additional = f"""
--> Example Product:
    > {example_product}
--> Applying whittaker to variables from sources:
    > `{'` and `'.join(whittaker)}`
--> Sharpening these variables:
    > `{'` and `'.join(sharpen)}`"""
        return base + additional

    @staticmethod
    def default_enhancers(prod, var):
        source = Configuration.source_func(prod)
        if "FILE:" in source:
            return []
        product_name = Configuration.pname_func(prod)
        mod = importlib.import_module(f"pywapor.collect.product.{source}")
        x = mod.default_post_processors(product_name, [var])[var]

        funcs = list()
        for x_ in x:
            if isinstance(x_, partial):
                funcs.append({
                    "func": f"{x_.func.__module__}.{x_.func.__name__}",
                    "args": x_.args,
                    "keywords": x_.keywords,
                })
            else:
                funcs.append({"func": f"{x_.__module__}.{x_.__name__}"})
        
        return funcs

    @classmethod
    def from_name(cls, name):
        log.info(f"--> Searching configuration for `{name}`.").add()

        folder = os.path.realpath(pywapor.__path__[0])
        available_configs = glob.glob(os.path.join(folder, "configs", "*.json"))
        available_names = sorted([os.path.splitext(os.path.split(x)[-1])[0] for x in available_configs])
        log.info(f"> Available configurations are `{'`, `'.join(available_names)}`")
    
        synonyms = {
            "level_1": "WaPOR2_level_1",
            "level_2": "WaPOR2_level_2",
            "level_3": "WaPOR2_level_3",
            "level_1_v2": "WaPOR2_level_1",
            "level_2_v2": "WaPOR2_level_2",
            "level_3_v2": "WaPOR2_level_3",
            "level_1_v3": "WaPOR3_level_1",
            "level_2_v3": "WaPOR3_level_2",
            "level_3_v3": "WaPOR3_level_3",
        }

        name_ = synonyms.get(name, name)

        selection = None
        for fh in available_configs:
            fn = os.path.splitext(os.path.split(fh)[-1])[0]
            if fn == name_:
                selection = fh
                break

        if isinstance(selection, type(None)):
            raise ValueError(f"No configuration found for name `{name}`, choose one from `{'`, `'.join(available_names)}`.")
        else:
            config = Configuration.from_json(selection)

        log.sub().info("--> Configuration set.")

        return config
    
    def to_json(self, fh):
        out = {
            "summary": self.summary, 
            "full": self.full, 
            "se_root": self.se_root, 
            "et_look": self.et_look,
        }
        class EncodeSetTuple(json.JSONEncoder):
            def encode(self, obj):
                def hint_tuples(item):
                    if isinstance(item, tuple):
                        return {'type': "tuple", 'value': list(item)}
                    if isinstance(item, list):
                        return [hint_tuples(e) for e in item]
                    if isinstance(item, set):
                        return {'type': "set", 'value': list(item)}
                    if isinstance(item, dict):
                        return {key: hint_tuples(value) for key, value in item.items()}
                    else:
                        return item
                return super(EncodeSetTuple, self).encode(hint_tuples(obj))
        
        json_string = json.dumps(out, cls = EncodeSetTuple)
        with open(fh, "w", encoding='utf8') as x:
            x.write(json_string)

    @classmethod
    def from_json(cls, fh):
        log.info(f"--> Loading configuration from `{os.path.split(fh)[1]}`.").add()
        def decode_set(dct):
            if dct.get("type", None) == "set":
                return set(dct.get('value'))
            if dct.get("type", None) == "tuple":
                return tuple(dct.get('value'))
            return dct
        with open(fh, "r", encoding='utf8') as fp:
            config_ = json.load(fp, object_hook=decode_set)
        
        config = cls(full = config_["full"], summary = config_["summary"], se_root = config_["se_root"], 
                 et_look = config_["et_look"])

        config.validate()
        log.sub().info("--> Configuration loaded.")
        return config
    
    @classmethod
    def from_summary(cls, summary):
        log.info("--> Creating configuration from summary.").add()

        valids = set(cls.variable_categories.keys()).union({"_EXAMPLE_", "_WHITTAKER_", "_ENHANCE_"})
        invalid_keys = set(summary.keys()).difference(valids)
        if invalid_keys:
            log.warning(f"--> Summary contains unrecognized keys ({invalid_keys}), they will be ignored.")
            summary = {k:v for k,v in summary.items() if k not in invalid_keys}

        example_product = summary.pop('_EXAMPLE_', '')
        temporal_interp = summary.pop('_WHITTAKER_', {})
        enhance = summary.pop('_ENHANCE_', {})

        full = dict()

        unique_enhancers = set([item for row in list(enhance.values()) for item in row])
        extra_vars = list()
        for enhancer in unique_enhancers:
            func = func_from_string(enhancer)
            extra_vars += [[(x,)] for x in func.__kwdefaults__["req_vars"]]

        for var_group in cls.et_look_vars + cls.se_root_vars + extra_vars:
            for var_ in var_group:
                for var in var_:

                    if var in full.keys():
                        continue

                    cat = cls.category_variables[var]
                    possible_prods = summary.get(cat, None)
                    
                    if isinstance(possible_prods, type(None)):
                        continue

                    products = []
                    t_interps = list()
                    for prod in possible_prods:
                        valid = cls.has_var(prod, var)
                        if valid:

                            enhancers = cls.default_enhancers(prod, var)

                            products += [{
                                "source": cls.source_func(prod), 
                                "product_name": cls.pname_func(prod), 
                                "enhancers": enhancers,
                                "is_example": prod == example_product,
                                }]
                            
                        x = temporal_interp.get(prod, None)
                        if not isinstance(x, type(None)):
                            t_interps += [x]
                    
                    if len(products) == 0:
                        continue

                    t_interps_unique = list(map(json.loads, set(map(lambda x: json.dumps(x, sort_keys=True), t_interps))))
                    # If no particular t_interp has been specified for this variable, use "linear".
                    if len(t_interps_unique) == 0:
                        temporal_interp_ = "linear"
                    # If a single unique t_interp has been specified, use that.
                    elif len(t_interps_unique) == 1:
                        temporal_interp_ = t_interps_unique[0]
                    # If multiple t_interp are specified for one variable, warn and use the first one.
                    else:
                        warn_string = str({"_WHITTAKER_": {x: temporal_interp[x] for x in possible_prods}})
                        log.warning(f"--> Multiple temporal interpolations specified for {var} (through `{warn_string}`), will continue using {t_interps_unique[0]}.")
                        temporal_interp_ = t_interps_unique[0]

                    variable_enhancers = enhance.get(var, [])

                    composite_type = "mean" if var.split("_")[-1] not in ["min", "max"] else var.split("_")[-1]

                    full[var] = {
                        "products": products,
                        "temporal_interp": temporal_interp_,
                        "variable_enhancers": variable_enhancers,
                        "spatial_interp": "bilinear",
                        "composite_type": composite_type,
                    }

        config = cls(full = full)
        config.validate()
        config.summarize()
        config.update_se_root_config()
        config.update_et_look_config()
        log.sub().info("--> Configuration created.")
        return config

    @staticmethod
    def has_var(prod, var, verbose = True):
        source = Configuration.source_func(prod)
        product_name = Configuration.pname_func(prod)
        if "FILE:" in source:
            if not verbose:
                log.info(f"> Variable `{var}` will be loaded from a file `{source}`.")
            return True
        mod = importlib.import_module(f"pywapor.collect.product.{source}")
        try:
            mod.default_vars(product_name, [var])
            valid = True
        except TypeError:
            valid = False
            if not verbose:
                log.warning(f"> {source}.{product_name} does not have a variable called `{var}`.") 
        return valid
    
    def validate(self):
        log.info("--> Validating configuration.").add()
        valids = []
        for var in self.full:
            for product in self.full[var]["products"]:
                prod = f"{product['source']}.{product['product_name']}"
                valid = Configuration.has_var(prod, var, verbose = False)
                valids.append(valid)
        if all(valids):
            log.info("> All specified variables sources are valid.")
        log.sub()
        return valid

    def update_se_root_config(self):
        log.info("--> Making configuration for SE_ROOT.").add()
        
        sharpened_vars = {var for var, config in self.full.items() if any(["sharpen" in x for x in config["variable_enhancers"]])}

        sharpeners = ["mndwi", "psri", "vari_red_edge", "bsi", "nmdi", "green", "nir"]
    
        se_root_config = dict()
        for var_group in self.se_root_vars:
            for var in var_group:
                if all([var_ in self.full.keys() for var_ in var]):
                    for var_ in var:
                        se_root_config[var_] = self.full[var_].copy()
                        _ = se_root_config[var_].pop("composite_type")
                    break
                elif var == var_group[-1]:
                    substring = [f"[`{'` and `'.join(x)}`]" for x in var_group]
                    log.warning(f"--> No configuration found for {' or '.join(substring)}.")
                else:
                    continue
        
        to_sharpen = set(sharpened_vars) & set(se_root_config)
        if to_sharpen:
            for var in sharpeners:
                if var in self.full.keys():
                    se_root_config[var] = self.full[var]
                else:
                    log.warning(f"--> No configuration found for `{var}` (required for sharpening of `{'` and `'.join(to_sharpen)}`).")

        self.se_root = se_root_config
        log.sub()

    def update_et_look_config(self):
        log.info("--> Making configuration for ET_LOOK.").add()
        
        sharpened_vars = {var for var, config in self.full.items() if any(["sharpen" in x for x in config["variable_enhancers"]])}

        sharpeners = ["mndwi", "psri", "vari_red_edge", "bsi", "nmdi", "green", "nir"]
    
        et_look_config = dict()
        for var_group in self.et_look_vars:
            for var in var_group:
                if all([var_ in self.full.keys() for var_ in var]):
                    for var_ in var:
                        et_look_config[var_] = self.full[var_].copy()
                    if len(var) == 0 and var == var_group[-1]:
                        substring = [f"[`{'` and `'.join(x)}`]" for x in var_group[:-1]]
                        log.info(f"--> No configuration found for {' or '.join(substring)}, will use constant value.")
                    break
                elif var == var_group[-1]:
                    substring = [f"[`{'` and `'.join(x)}`]" for x in var_group]
                    log.warning(f"--> No configuration found for {' or '.join(substring)}.")
                else:
                    continue
        
        to_sharpen = set(sharpened_vars) & set(et_look_config)
        if to_sharpen:
            for var in sharpeners:
                if var in self.full.keys():
                    et_look_config[var] = self.full[var]
                else:
                    log.warning(f"--> No configuration found for `{var}` (required for sharpening of `{'` and `'.join(to_sharpen)}`).")

        self.et_look = et_look_config
        log.sub()

    def summarize(self):
        log.info("--> Making summary of configuration.")
        
        summary = dict()
        for var, config in self.full.items():
            sources = set()
            for x in config["products"]:
                if "FILE:" in x["source"]:
                    sources.add(x["source"])
                else:
                    sources.add(f'{x["source"]}.{x["product_name"]}')
            cat = self.category_variables.get(var, "unknown")
            if cat == "other":
                log.warning(f"--> Unknown variable `{var}` found.")
            if cat in list(summary.keys()):
                summary[cat] = summary[cat].union(sources)
            else:
                summary[cat] = sources

        whittaker = dict()
        enhance = dict()
        example = None
        for k,v in self.full.items():
            if isinstance(example, type(None)):
                example_ = [f"{prod['source']}.{prod['product_name']}" for prod in v["products"] if prod["is_example"]]
                if len(example_) > 0:
                    example = example_[0]
            if isinstance(v["temporal_interp"], dict):
                if v["temporal_interp"].get("method", "") == "whittaker":
                    for prod in v["products"]:
                        whittaker[f"{prod['source']}.{prod['product_name']}"] = v["temporal_interp"]
            if len(v["variable_enhancers"]) > 0:
                enhance[k] = v["variable_enhancers"]

        summary["_WHITTAKER_"] = whittaker
        summary["_ENHANCE_"] = enhance
        summary["_EXAMPLE_"] = example

        self.summary = summary

class Project():

    def __init__(self, project_folder, bb, period, configuration = None):

        assert bb[::2][0] < bb[::2][1], "Invalid Bounding-Box"
        self.lonlim = bb[::2]
        assert bb[1::2][0] < bb[1::2][1], "Invalid Bounding-Box"
        self.latlim = bb[1::2]
        if not os.path.isdir(project_folder):
            os.makedirs(project_folder)
        self.folder = project_folder
        self.period = adjust_timelim_dtype(period)
        self.configuration = configuration
        self.dss = None
        self.bb = bb

        os.environ["pyWaPOR_bb"] = str(self.bb)
        os.environ["pyWaPOR_period"] = str(self.period)

        warnings.filterwarnings("ignore", message="invalid value encountered in power")

        adjust_logger(True, self.folder, "INFO")

        if os.path.isfile(os.path.join(self.folder, "se_root_in.nc")):
            self.se_root_in = open_ds(os.path.join(self.folder, "se_root_in.nc"))
        else:
            self.se_root_in = None

        se_root_outs = glob.glob(os.path.join(self.folder, "se_root_out*.nc"))
        if se_root_outs:
            self.se_root_out = open_ds(max(se_root_outs, key=os.path.getmtime))
        else:
            self.se_root_out = None

        if os.path.isfile(os.path.join(self.folder, "et_look_in.nc")):
            self.et_look_in = open_ds(os.path.join(self.folder, "et_look_in.nc"))
        else:
            self.et_look_in = None

        et_look_outs = glob.glob(os.path.join(self.folder, "et_look_out*.nc"))
        if et_look_outs:
            self.et_look_out = open_ds(max(et_look_outs, key=os.path.getmtime))
        else:
            self.et_look_out = None

        log.info("> PROJECT").add()
        log.info(self.__repr__())
        self.check_pywapor_version()
        self.check_gdal_drivers()
        log.sub().info("< PROJECT")

    def check_gdal_drivers(self):
        required_drivers = [
            "GTiff",
            "JP2OpenJPEG",
            "NETCDF",
            "HDF5",
        ]
        log.info(f"--> GDAL ({gdal.__version__}):").add()
        drivers_check = True
        for driver in required_drivers:
            if isinstance(gdal.GetDriverByName(driver), type(None)):
                log.warning(f"> Driver `{driver}` not found.")
                drivers_check = False
        
        if drivers_check:
            log.info("> All required GDAL drivers found.")
        log.sub()

    def check_pywapor_version(self):
        current_version = pywapor.__version__
        package = 'pywapor'
        log.info(f"--> pyWaPOR ({current_version}):").add()
        try:
            response = requests.get(f'https://pypi.org/pypi/{package}/json')
            response.raise_for_status()
        except Exception as _:
            log.warning("> Unable to check for pyWaPOR updates.")
        else:
            latest_version = response.json()['info']['version']
            if latest_version == current_version:
                log.info("> Up to date.")
            else:
                log.warning(f"> Latest version is '{latest_version}'.")
                log.warning("> Please update pywapor.")
        log.sub()

    def __repr__(self):
        project_str = f"""--> Project Folder:
        > {self.folder}
    --> Period:
        > {self.period[0]} - {self.period[1]}
    --> Bounding-Box:\n
                 {self.latlim[1]:8.4f}
                 ┌─────────┐
                 │         │
       {self.lonlim[0]:9.4f} │         │{self.lonlim[1]:9.4f}
                 │         │
                 └─────────┘
                 {self.latlim[0]:8.4f}\n
    --> Configuration:
        > {self.configuration}"""
        return project_str

    def load_configuration(self, name = None, summary = None, json = None):
        if not isinstance(name, type(None)) and not isinstance(summary, type(None)):
            raise ValueError("Only one of `level` and `summary` can be specified.")
        log.info("> CONFIGURATION").add()
        if not isinstance(name, type(None)):
            self.configuration = Configuration.from_name(name)
        elif not isinstance(summary, type(None)):
            summary = summary.copy()
            self.configuration = Configuration.from_summary(summary)
        elif not isinstance(json, type(None)):
            self.configuration = Configuration.from_json(json)
        else:
            raise ValueError("At least one of `level` or `summary` needs to be specified.")
        
        log.info("--> Summary:").add()
        summary_str = self.configuration.__repr__()

        for line in summary_str.split("\n"):
            log.info(line)

        log.sub().sub().info("< CONFIGURATION")
        return self.configuration
    
    def validate_project_folder(self):
        log.info("--> Checking project folder.").add()
        if not os.path.isabs(self.folder):
            self.folder = os.path.abspath(self.folder)
            log.warning(f"--> Consider specifying an absolute path to your project folder, continueing with `{self.folder}`.")
        if not os.path.isdir(self.folder):
            log.info(f"--> Folder does not exist, creating a new folder at `{self.folder}`.")
            os.makedirs(self.folder)
        dir_list = os.listdir(self.folder)
        bb = [self.lonlim[0], self.latlim[0], self.lonlim[1], self.latlim[1]]
        check = True

        if "log.txt" in dir_list:
            log_file = os.path.join(self.folder, "log.txt")
            project_folders, periods, bbs, temp_files = Project.parse_log_file(log_file)

            if len(set(project_folders)) > 1:
                log.info("--> Project folder has previously been moved.")
            else:
                ... # all good.

            if len(set(bbs)) > 1:
                log.warning("--> You've previously defined different bounding-boxes inside this project folder, this can result in unexpected behaviour.")
                check = False
            elif len(set(bbs)) == 1:
                bb_consistent = np.all([np.isclose(x, y) for x,y in zip(bbs[0], bb)])
                if not bb_consistent:
                    log.warning("--> You've changed the bounding-box of your project without changing the project folder, this can result in unexpected behaviour.")
                    check = False
            else:
                ... # all good.

            period_ = tuple([np.datetime64(x) for x in self.period])
            if len(set(periods)) > 1:
                log.warning("--> You've previously defined different periods inside this project folder, this can result in unexpected behaviour.")
                check = False
            elif len(set(periods)) == 1:
                period_consistent = period_ == periods[0]
                if not period_consistent:
                    log.warning("--> You've changed the period of your project without changing the project folder, this can result in unexpected behaviour.")
                    check = False
            else:
                ... # all good.

            if len(dir_list) > 1:
                self.clean_project_folder()

        elif dir_list:
            log.warning("--> Consider specifying an empty project folder.")
            check = False
        else:
            ... # all good.

        if check:
            log.info("--> All good.")

        log.sub()

    def clean_project_folder(self):

        n_corrupt = n_wrong = n_unkown = n_good = n_temp = 0
        to_remove = list()
        fhs = glob.glob(os.path.join(self.folder, "**/*.nc"), recursive=True)
        log_file = os.path.join(self.folder, "log.txt")

        if os.path.isfile(log_file):
            temp_files = Project.parse_log_file(log_file)[-1]
            for fh in temp_files:
                if fh in fhs:
                    fhs.remove(fh)
                if os.path.isfile(fh):
                    to_remove.append(fh)
                    n_temp += 1

        log.info(f"--> Checking {len(fhs)} files.").add()

        for fh in fhs.copy():
            corrupt = is_corrupt_or_empty(fh)
            if corrupt:
                to_remove.append(fh)
                fhs.remove(fh)
                n_corrupt += 1
            
        for fh in fhs:
            wrong = has_wrong_bb_or_period(fh, self.bb, self.period)
            if wrong:
                to_remove.append(fh)
                n_wrong += 1
            elif isinstance(wrong, type(None)):
                n_unkown += 1
            else:
                n_good += 1

        if n_temp > 0:
            log.info(f"--> Found {n_temp} temporary files that can be removed.")
        log.info(f"--> Found {n_good} files in good condition.")
        if n_corrupt > 0:
            log.warning(f"--> Found {n_corrupt} corrupt or empty files.")
        if n_wrong > 0:
            log.warning(f"--> Found {n_wrong} files with an incorrect `bb` or `period`.")
        if n_unkown > 0:
            log.info(f"--> Couldn't determine state of {n_unkown} files.").sub()
        
        if n_corrupt + n_wrong + n_temp > 0:
            response = input(prompt = f"Remove {n_corrupt} corrupt, {n_wrong} wrong and {n_temp} temporary files? (Yes/No)")
            if response in ["Y", "y", "Yes", "yes"]:
                for x in to_remove:
                    remove_ds(x)

    @staticmethod
    def parse_log_file(log_file):

        with open(log_file, "r", encoding='utf8') as f:
            log_string = f.read()

        project_folders = re.findall(r"--> Project Folder:\n        > (.*)", log_string)
        
        periods_ = re.findall(r"--> Period:\n        > (.*) - (.*)", log_string)
        periods = [tuple([np.datetime64(x) for x in period]) for period in periods_]

        bbs_ = re.findall(r"--> Bounding-Box:\n\n\s*(\d{1,2}.\d{4})\n.*\n.*\n\s*(\d{1,2}.\d{4})\D*(\d{1,2}.\d{4})\n.*\n.*\n\D*(\d{1,2}.\d{4})", log_string)
        bbs = [tuple([[float(x) for x in bb][i] for i in [1,3,2,0]]) for bb in bbs_]
        
        temp_files = re.findall(r"--> Unable to delete temporary file `(.*)`", log_string)

        return project_folders, periods, bbs, temp_files

    def set_passwords(self):

        log.info("> PASSWORDS").add()

        req_accounts = {
            "MODIS": "NASA",
            "SRTM": "NASA",
            "MERRA2": "NASA",
            "TERRA": "TERRA",
            "ERA5": "CDS",
            "LANDSAT": "EARTHEXPLORER",
            "SENTINEL2": "COPERNICUS_DATA_SPACE",
            "SENTINEL3": "COPERNICUS_DATA_SPACE",
        }

        all_accounts = list()
        for v in self.configuration.summary.values():
            if isinstance(v, set):
                for source_product in v:
                    name = Configuration.source_func(source_product)
                    account = req_accounts.get(name, None)
                    if not isinstance(account, type(None)):
                        all_accounts.append(account)

        x = set(all_accounts)
        log.info(f"--> Accounts needed for `{'`, `'.join(x)}`.").add()

        for account in x:
            _ = pywapor.collect.accounts.get(account)

        log.sub().info("--> All set!")
        log.sub().info("< PASSWORDS")

    def download_data(self, buffer_timelim = True):
        assert not isinstance(self.configuration, type(None)), "Please load a configuration before continueing."

        log.info("> DOWNLOADER").add()

        example_t_vars = [x for x in ["lst", "bt"] if x in self.configuration.full.keys()]
        example_sources = {k:v for k,v in self.configuration.full.items() if k in example_t_vars}
        other_sources = {k:v for k,v in self.configuration.full.items() if k not in example_t_vars}

        if buffer_timelim:
            bins = time_bins(self.period, 1)
            adjusted_timelim = [bins[0], bins[-1]]
            buffered_timelim = [adjusted_timelim[0] - np.timedelta64(3, "D"), 
                                adjusted_timelim[1] + np.timedelta64(3, "D")]
        else:
            adjusted_timelim = self.period
            buffered_timelim = self.period

        example_dss, _ = collect_sources(
                            self.folder,
                            example_sources,
                            self.latlim,
                            self.lonlim,
                            adjusted_timelim,
                            landsat_order_only = True
                            )
        
        other_dss, _ = collect_sources(
                            self.folder,
                            other_sources,
                            self.latlim,
                            self.lonlim,
                            buffered_timelim,
                            )
        
        # If there are example-t variables that rely on landsat, try one more time to collect them.
        if np.any(list({var: np.any([product_info["source"] == "LANDSAT" for product_info in info["products"]]) for var, info in example_sources.items()}.values())):
            example_dss, _ = collect_sources(
                                            self.folder, 
                                            example_sources, 
                                            self.latlim, 
                                            self.lonlim, 
                                            adjusted_timelim
                                            )

        self.dss = {**example_dss, **other_dss}
        
        log.sub().info("< DOWNLOADER")
    
        return self.dss
    
    def run_pre_se_root(self, forced = False):
        if isinstance(self.se_root_in, type(None)) or forced:
            self.se_root_in = pywapor.pre_se_root.main(self.folder, self.latlim, self.lonlim, self.period, sources = self.configuration.se_root)
        else:
            log.info("> PRE_SE_ROOT").add()
            log.info(f"--> Re-using `{os.path.split(self.se_root_in.encoding['source'])[-1]}`.")
            log.sub().info("< PRE_SE_ROOT")
        return self.se_root_in

    def run_se_root(self, se_root_version = "v3", export_vars = "default", chunks = {"time": -1, "x": 500, "y": 500}):
        self.se_root_out = pywapor.se_root.main(
                                                self.se_root_in, 
                                                se_root_version = se_root_version, 
                                                export_vars = export_vars, 
                                                chunks = chunks
                                                )
        return self.se_root_out
    
    def run_pre_et_look(self, forced = False):
        if isinstance(self.et_look_in, type(None)) or forced:
            self.et_look_in = pywapor.pre_et_look.main(self.folder, self.latlim, self.lonlim, self.period, sources = self.configuration.et_look)
        else:
            log.info("> PRE_ET_LOOK").add()
            log.info(f"--> Re-using `{os.path.split(self.et_look_in.encoding['source'])[-1]}`.")
            log.sub().info("< PRE_ET_LOOK")
        return self.et_look_in

    def run_et_look(self, et_look_version = "v3", export_vars = "default", chunks = {"time_bins": -1, "x": 500, "y": 500}):
        self.et_look_out = pywapor.et_look.main(
                                                self.et_look_in, 
                                                et_look_version = et_look_version, 
                                                export_vars = export_vars, 
                                                chunks = chunks
                                                )
        return self.et_look_out

if __name__ == "__main__":

    timelim = ["2023-11-01", "2023-11-30"]
    latlim = [21.9692194682626933, 21.9939120838340507]
    lonlim = [91.9371349243682801, 91.9657566608824339]
    bb = [91.9371349243682801, 21.9692194682626933, 91.9657566608824339, 21.9939120838340507]
    project_folder = r"/Users/hmcoerver/Local/pywapor_bgdX"

    adjust_logger(True, project_folder, "INFO")

    all_configs = {
        "WaPOR2_level_1",
        "WaPOR2_level_2",
        "WaPOR2_level_3",
        "WaPOR2_level_1",
        "WaPOR2_level_2",
        "WaPOR2_level_3",
        "WaPOR3_level_2",
        "WaPOR3_level_3",
        "nrt",
        "all_in",
    }

    project = pywapor.Project(project_folder, bb, timelim)

    # project.validate_project_folder()

    # project.load_configuration(name = "WaPOR3_level_2")

    summary = {
        # Define which products to use.
        'elevation': {},
        'meteorological': {},
        'optical': {'LANDSAT.LC09_SR'},
        'precipitation': {},
        'solar radiation': {},
        'statics': {},
        'thermal': {},
        'soil moisture': {},

        # Define which product to reproject the other products to.
        '_EXAMPLE_': 'LANDSAT.LC09_SR', 

        # # Define any special functions to apply to a specific variable.
        # '_ENHANCE_': {"bt": ["pywapor.enhancers.dms.thermal_sharpener.sharpen"],},

        # # Choose which products should be gapfilled.
        # '_WHITTAKER_': {
        #     'SENTINEL2.S2MSI2A_R20m': {'lmbdas': 1000.0, 'method': 'whittaker'}, 
        #     'VIIRSL1.VNP02IMG': {'a': 0.85, 'lmbdas': 1000.0, 'method': 'whittaker'}},
        }
    
    project.load_configuration(summary = summary)



    project.set_passwords()
    dss = project.download_data()
    
    # se_root_in = project.run_pre_se_root()
    # se_root = project.run_se_root()
    # et_look_in = project.run_pre_et_look()
    # et_look = project.run_et_look()








import numpy as np
import xarray as xr

def atmospheric_canopy_resistance(lai_eff, stress_rad, stress_vpd,
                                  stress_temp, rs_min=70, rcan_max=1000000.):
    r"""
    Computes canopy resistance excluding soil moisture stress.

    .. math ::

        r_{canopy,0}=\left(\frac{r_{s,min}}{I_{lai, eff}}\right)\cdot
        \left(\frac{1}{S_{T} \cdot S_{V} \cdot S_{r}}\right)

    Parameters
    ----------
    lai_eff : float
        effective leaf area index, 
        :math:`I_{lai,eff}`
        [-]
    stress_temp : float
        stress factor for air temperature, 
        :math:`S_{t}`
        [-]
    stress_vpd : float
        stress factor for vapour pressure deficit, 
        :math:`S_{v}`
        [-]
    stress_rad : float
        stress factor for radiation, 
        :math:`S_{r}`
        [-]
    rs_min : float
        Minimal stomatal resistance, 
        :math:`r_{s_min}`
        [sm-1]
    rcan_max : float
        Maximum stomatal resistance, 
        :math:`r_{can_max}`
        [sm-1]

    Returns
    -------
    r_canopy_0 : float
        atmospheric canopy resistance, 
        :math:`r_{canopy,0}`
        [sm-1]

    Examples
    --------
    >>> import ETLook.resistance as res
    >>> res.atmospheric_canopy_resistance(0.9, 0.4, 0.9, 0.94)
    229.839768846861
    """

    bulk_stress = stress_rad * stress_temp * stress_vpd
    max_mask = np.logical_or(bulk_stress == 0, lai_eff == 0)

    r_canopy_0 = (rs_min / lai_eff) / bulk_stress
    if isinstance(r_canopy_0, xr.DataArray):
        r_canopy_0 = xr.where(max_mask, rcan_max, r_canopy_0)
    else:
        r_canopy_0[max_mask] = rcan_max

    return r_canopy_0


def canopy_resistance(r_canopy_0, stress_moist, rcan_max=1000000.):
    r"""
    Computes canopy resistance.

    .. math ::

        r_{canopy}=\frac{r_{canopy,0}}{S_{m}}

    Parameters
    ----------
    r_canopy_0 : float
        Atmospheric canopy resistance, 
        :math:`r_{canopy_0}`
        [sm-1]
    stress_moist : float
        stress factor for root zone soil moisture, 
        :math:`S_{m}`
        [-]
    rcan_max : float
        Maximum stomatal resistance, 
        :math:`r_{can_max}`
        [sm-1]

    Returns
    -------
    r_canopy : float
        canopy resistance, 
        :math:`r_{canopy}`
        [sm-1]

    Examples
    --------
    >>> import ETLook.resistance as res
    >>> res.canopy_resistance(218, 0.8)
    272.5
    """

    if isinstance(r_canopy_0, xr.DataArray):
        r_canopy = xr.where(stress_moist == 0, rcan_max, r_canopy_0/stress_moist)
    else:
        r_canopy = np.where(stress_moist == 0, rcan_max, r_canopy_0/stress_moist)

    return r_canopy


def soil_resistance(se_top, land_mask=1, r_soil_pow=-2.1, r_soil_min=800):
    r"""
    Computes soil resistance.

    .. math ::

        r_{soil}=r_{soil, min}\cdot \left(S_{e,top}\right)^{a}

    Parameters
    ----------
    r_soil_min : float
        Minimum soil resistance, 
        :math:`r_{soil,min}`
        [sm-1]
    se_top : float
        Top soil effective saturation, 
        :math:`S_{e,top}`
        [-]
    r_soil_pow : float
        Power soil resistance function, 
        :math:`a`
        [-]
    land_mask : int
        land use classification, 
        :math:`l`
        [-]

    Returns
    -------
    r_soil : float
        soil resistance, 
        :math:`r_{soil}`
        [sm-1]

    Examples
    --------
    >>> import ETLook.resistance as res
    >>> res.soil_resistance(se_top=0.9)
    998.1153098304111
    """
    if isinstance(land_mask, xr.DataArray):
        res = xr.ones_like(land_mask) * r_soil_min * se_top ** r_soil_pow
        res = xr.where(land_mask == 2, 0, res)
    else:
        res = np.ones(land_mask.shape) * r_soil_min * se_top ** r_soil_pow
        res[land_mask == 2] = 0

    return res



__all__ = [
    'variables', 
    'processing_functions', 
    'compositer', 
    'pre_defaults', 
    'lazifier', 
    'reproject', 
    'levels', 
    'log_indenter'
    ]
from . import variables, processing_functions, compositer, pre_defaults, lazifier, reproject, levels, log_indenter


from setuptools import setup, find_packages

setup(
    name = 'pywapor',
    version = '3.5.16',
    url = 'https://www.fao.org/aquastat/py-wapor/',
    author = "FAO",
    author_email = "bert.coerver@fao.org",
    license = "Apache",
    packages = find_packages(include = ['pywapor', 'pywapor.*']),
    include_package_data=True,
    python_requires='>=3.10',
    install_requires = [
        # 'libgdal-netcdf',
        # 'libgdal-jp2openjpeg',
        'netcdf4',
        'gdal>=3,<4.0',
        'xarray',
        'numpy<2',
        'pandas<3',
        'requests',
        'matplotlib',
        'pyproj',
        'scipy',
        'pycurl',
        'joblib',
        'bs4',
        'rasterio',
        'bottleneck',
        'tqdm',
        'dask',
        'rioxarray>=0.15.5',
        'cryptography',
        'cachetools',
        'cdsapi>=0.7.0',
        'shapely',
        'lxml',
        'scikit-learn',
        'numba',
        'xmltodict',
    ],
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ],
)


"""Importing `log` from this module starts a logging handler, which can be logged to 
with `log.info("log this info")` etc.
"""
from pywapor.general.log_indenter import IndentedLoggerAdapter
import logging
import os
from tqdm.contrib.logging import logging_redirect_tqdm
import warnings
import sys

class MyCoolFormatter(logging.Formatter):
    """
    See https://jupyterbook.org/en/stable/content/code-outputs.html#ansi-outputs
    for colors. E.g. below:
    30 gives black font (and 1;30 would give bold-black font)
    41 gives red background.
    """
    def format(self, record):
        if record.levelno > logging.INFO:
            self._style._fmt = "\x1b[30;41m%(message)s\x1b[0m"
        else:
            self._style._fmt = "%(message)s"
        return super().format(record)

log_settings = logging.getLogger(__name__)
__handler__ = logging.StreamHandler(stream=sys.stdout)
__handler__.setFormatter(MyCoolFormatter())
__handler__.setLevel("INFO")
log_settings.addHandler(__handler__)
log = IndentedLoggerAdapter(log_settings)

def adjust_logger(log_write, folder, log_level, testing = False):
    """Function to adjust the default logging settings that get initiated by
    importing this module.

    Parameters
    ----------
    log_write : bool
        Stop or start writing to log.
    folder : str
        Path to folder in which to store `"log.txt"`.
    log_level : str
        Set the log level.
    """
    if log_write and any([isinstance(x, logging.FileHandler) for x in log_settings.handlers]):
        handlers_to_remove = [x for x in log_settings.handlers if isinstance(x, logging.FileHandler)]
        for handler in handlers_to_remove:
            log_settings.removeHandler(handler)

    if log_write and not any([isinstance(x, logging.FileHandler) for x in log_settings.handlers]):
        if not os.path.isdir(folder):
            os.makedirs(folder)
        handler = logging.FileHandler(filename = os.path.join(folder, "log.txt"), encoding="utf-8")
        formatter = logging.Formatter('{asctime} {levelname:>9s}: {message}', style='{')
        handler.setFormatter(formatter)
        handler.setLevel("DEBUG")
        log_settings.addHandler(handler)
        logging_redirect_tqdm(loggers = log_settings.handlers)
        log_settings.propagate = False

    if testing:
        log_settings.propagate = True
        def fancy_warning(message):
            IndentedLoggerAdapter(log_settings).warning(message)
            warnings.warn(message, stacklevel=2)
        log.warning = fancy_warning

    log_settings.setLevel(log_level)

    if testing:
        from osgeo import gdal
        log.info(f"--> Using GDAL ({gdal.__version__}) from `{gdal.__file__}`.")

if __name__ == "__main__":

    adjust_logger(True, r"/Users/hmcoerver/Local/cog_test", "INFO")

    log.info("tst").add().info("tralala").sub()
    log.warning("balblblblasdas").add()
    log.info("hihihi").add()
    log.sub()
    log.warning("no").add()
    log.sub().sub().info("yes")


"""
    The meteo module contains all functions related to meteorological
    variables. All meteorological functions can be calculated on a daily
    or instantaneous basis. Base functions are available also. The daily
    functions have a 'daily' extension, instantaneous functions have a 'inst'
    extension.

"""
from pywapor.et_look_v2_v3 import constants as c
import numpy as np
import xarray as xr

def mean_temperature_kelvin_daytime(t_air_k_min, t_air_k_max):
    r"""
    Computes the mean temperature over the daily sunshine period.

    Parameters
    ----------
    t_air_k_min : float
        maximum air temperature
        :math:`T_{a,min}`
        [K]
    t_air_k_max : float
        maximum air temperature
        :math:`T_{a,max}`
        [K]

    Returns
    -------
    t_air_k_12 : float
        daytime air temperature
        :math:`T_{a,12}`
        [K]

    """
    return 0.25 * t_air_k_min + 0.75 * t_air_k_max


def air_temperature_kelvin(t_air):
    r"""
    Converts air temperature from Celcius to Kelvin, where 0 degrees Celcius
    is equal to 273.15 degrees Kelvin.

    Parameters
    ----------
    t_air : float
        air temperature, 
        :math:`T_a`
        [C]

    Returns
    -------
    t_air_k : float
        air temperature, 
        :math:`T_a`
        [K]

    Examples
    --------
    >>> from ETLook import meteo
    >>> meteo.air_temperature_kelvin(12.5)
    285.65
    """
    return t_air + c.zero_celcius


def air_temperature_kelvin_daily(t_air_24):
    """Like :func:`air_temperature_kelvin` but as a daily average.

    Parameters
    ----------
    t_air_24 : float
        daily air temperature, 
        :math:`T_{a,24}`
        [C]

    Returns
    -------
    t_air_k_24 : float
        daily air temperature, 
        :math:`T_{a,24}`
        [K]
    """
    return air_temperature_kelvin(t_air_24)


def air_temperature_kelvin_inst(t_air_i):
    """Like :func:`air_temperature_kelvin` but as an instantaneous value.

    Parameters
    ----------
    t_air_i : float
        instantaneous air temperature, 
        :math:`T_{a,i}`
        [C]

    Returns
    -------
    t_air_k_i : float
        instantaneous air temperature, 
        :math:`T_{a,i}`
        [K]
    """
    return air_temperature_kelvin(t_air_i)


def wet_bulb_temperature_kelvin_inst(t_wet_i):
    """
    Converts wet bulb temperature from Celcius to Kelvin, where 0
    degrees Celcius is equal to 273.15 degrees Kelvin.

    Parameters
    ----------
    t_wet_i : float
        instantaneous wet bulb temperature, 
        :math:`T_{w,i}`
        [C]

    Returns
    -------
    t_wet_k_i : float
        instantaneous wet bulb temperature, 
        :math:`T_{w,i}`
        [K]
    """
    return air_temperature_kelvin(t_wet_i)


def disaggregate_air_temperature(t_air_coarse, z, z_coarse, lapse=-0.006):
    r"""
    Disaggregates GEOS or MERRA or another coarse scale air temperature using
    two digital elevation models. One DEM for the target resolution, another
    DEM smoothed from the original air temperature resolution to the target
    resolution.

    .. math ::

        T_{a}=T_{a,c}+(z-z_{c}) \cdot L_{T}-T_{K,0}

    where the following constant is used

    * :math:`T_{K,0}` = 273.15 K is equal to 0 degrees Celsius

    Parameters
    ----------
    t_air_coarse : float
        air temperature at coarse resolution, 
        :math:`T_{a,c}`
        [K]
    z : float
        elevation, 
        :math:`z`
        [m]
    z_coarse : float
        elevation at coarse resolution, 
        :math:`z_{c}`
        [m]
    lapse : float
        lapse rate, 
        :math:`L_{T}`
        [K m-1]

    Returns
    -------
    t_air : float
        air temperature, 
        :math:`T_{a}`
        [C]

    Notes
    -----
    The input air temperature is specified in Kelvin. The output air
    temperature is specified in C.

    Examples
    --------
    >>> from ETLook import meteo
    >>> meteo.disaggregate_air_temperature(24.5+273.15, 10, 5)
    24.47
    """
    return t_air_coarse + ((z - z_coarse) * lapse) - c.zero_celcius


def disaggregate_air_temperature_daily(t_air_24_coarse, z, z_coarse, lapse=-0.006):
    r"""Like :func:`disaggregate_air_temperature` but as a daily average.

    Parameters
    ----------
    t_air_24_coarse : float
        daily air temperature at coarse resolution, 
        :math:`T_{a,24,c}`
        [K]
    z : float
        elevation, 
        :math:`z`
        [m]
    z_coarse : float
        elevation at coarse resolution, 
        :math:`z_{c}`
        [m]
    lapse : float
        lapse rate, 
        :math:`L`
        [K m-1]

    Returns
    -------
    t_air_24 : float
        daily air temperature, 
        :math:`T_{a,24}`
        [C]

    Notes
    -----
    The input air temperature is specified in Kelvin. The output air
    temperature is specified in C.
    """
    return disaggregate_air_temperature(t_air_24_coarse, z, z_coarse, lapse)


def disaggregate_air_temperature_inst(t_air_i_coarse, z, z_coarse, lapse=-0.006):
    r"""Like :func:`disaggregate_air_temperature` but as a instantaneous value.

    Parameters
    ----------
    t_air_i_coarse : float
        instantaneous air temperature at coarse resolution, 
        :math:`T_{a,i,c}`
        [K]
    z : float
        elevation, 
        :math:`z`
        [m]
    z_coarse : float
        elevation at coarse resolution, 
        :math:`z_{c}`
        [m]
    lapse : float
        lapse rate, 
        :math:`L`
        [K m-1]

    Returns
    -------
    t_air_i : float
        instantaneous air temperature, 
        :math:`T_{a,i}`
        [C]

    Notes
    -----
    The input air temperature is specified in Kelvin. The output air
    temperature is specified in C.
    """
    return disaggregate_air_temperature(t_air_i_coarse, z, z_coarse, lapse)


def disaggregate_dew_point_temperature_inst(t_dew_coarse_i, z, z_coarse, lapse_dew=-0.002):
    r"""
    Disaggregates geos dew point temperature using lapse rate and difference between
    smoothed coarse scale DEM and fine scale DEM.

    Parameters
    ----------
    t_dew_coarse_i : float
        coarse instantaneous dew point temperature, 
        :math:`T_{dew,coarse}`
        [C]
    z : float
        elevation, 
        :math:`z`
        [m]
    z_coarse : float
        smoothed elevation at coarse resolution, 
        :math:`z`
        [m]
    lapse_dew : float
        lapse rate, 
        :math:`L`
        [K m-1]

    Returns
    -------
    t_dew_i : float
        instantaneous dew point temperature, 
        :math:`T_{dew,i}`
        [C]
    """
    return t_dew_coarse_i + ((z - z_coarse) * lapse_dew)


def vapour_pressure_from_specific_humidity(qv, p_air):
    r"""
    Computes the vapour pressure :math:`e_a` in [mbar] using specific humidity
    and surface pressure.

    .. math ::

        e_{a}=\frac{q_{v} \cdot P}{\varepsilon}

    where the following constant is used

    * :math:`\varepsilon` = ratio of molecular weight of water to
      dry air = 0.622 [-]


    Parameters
    ----------
    qv : float
        specific humidity, 
        :math:`q_{v}`
        [kg/kg]
    p_air : float
        air pressure, 
        :math:`P`
        [mbar]

    Returns
    -------
    vp : float
        vapour pressure, 
        :math:`e_{a}`
        [mbar]
    """
    return (qv * p_air) / c.r_mw

def specific_humidity_from_vapour_pressure(vp, p_air):
    r"""
    Computes specific humidity using te vapour pressure :math:`e_a` in [mbar]
    and surface pressure.

    .. math ::

        e_{a}=\frac{q_{v} \cdot P}{\varepsilon}

    where the following constant is used

    * :math:`\varepsilon` = ratio of molecular weight of water to
      dry air = 0.622 [-]


    Parameters
    ----------
    vp : float
        vapour pressure, 
        :math:`e_{a}`
        [mbar]
    p_air : float
        air pressure, 
        :math:`P`
        [mbar]

    Returns
    -------
    qv : float
        specific humidity, 
        :math:`q_{v}`
        [kg/kg]
    """
    return (vp * c.r_mw) / p_air

def vapour_pressure_from_specific_humidity_daily(qv_24, p_air_24):
    """Like :func:`vapour_pressure_from_specific_humidity` but as a daily average.

    Parameters
    ----------
    qv_24 : float
        daily specific humidity, 
        :math:`q_{v,24}`
        [kg/kg]
    p_air_24 : float
        daily air pressure, 
        :math:`P_{24}`
        [mbar]

    Returns
    -------
    vp_24 : float
        daily vapour pressure, 
        :math:`e_{a,24}`
        [mbar]
    """
    return vapour_pressure_from_specific_humidity(qv_24, p_air_24)


def vapour_pressure_from_dewpoint(t_dew):
    r"""
    .. math ::
        e_{a}=6.108\exp\left[\frac{17.27T_{d}}{T_{d}+237.3}\right]

    Parameters
    ----------
    t_dew : float
        dewpoint temperature, 
        :math:`T_d`
        [°C]

    Returns
    -------
    vp : float
        vapour pressure, 
        :math:`e_a`
        [mbar]

    Examples
    --------
    >>> from ETLook import meteo
    >>> meteo.vapour_pressure_from_dewpoint(20)
    23.382812709274457
    """

    return saturated_vapour_pressure(t_dew)


def vapour_pressure_from_dewpoint_daily(t_dew_24):
    r"""
    .. math ::
        e_{a}=6.108\exp\left[\frac{17.27T_{d,24}}{T_{d,24}+237.3}\right]

    Parameters
    ----------
    t_dew_24 : float
        dewpoint temperature, 
        :math:`T_d`
        [°C]

    Returns
    -------
    vp_24 : float
        vapour pressure, 
        :math:`e_{a,24}`
        [mbar]

    Examples
    --------
    >>> from ETLook import meteo
    >>> meteo.vapour_pressure_from_dewpoint_daily(20)
    23.382812709274457
    """

    return saturated_vapour_pressure(t_dew_24)


def vapour_pressure_from_dewpoint_inst(t_dew_i):
    r"""
    .. math ::
        e_{a,i}=6.108\exp\left[\frac{17.27T_{d,i}}{T_{d,i}+237.3}\right]

    Parameters
    ----------
    t_dew_i : float
        instantaneous dew point temperature, 
        :math:`T_{dew,i}`
        [°C]

    Returns
    -------
    vp_i : float
        vapour pressure, 
        :math:`e_{a,i}`
        [mbar]

    Examples
    --------
    >>> from ETLook import meteo
    >>> meteo.vapour_pressure_from_dewpoint_inst(20)
    23.382812709274457
    """

    return saturated_vapour_pressure(t_dew_i)


def saturated_vapour_pressure_minimum(t_air_min_coarse):
    """Like :func:`saturated_vapour_pressure` but based on daily minimum air temperature. This
    is only relevant for reference ET calculations.

    Parameters
    ----------
    t_air_min_coarse : float
        daily minimum air temperature, 
        :math:`T_{a,min}`
        [C]

    Returns
    -------
    svp_24_min : float
        daily saturated vapour pressure based on minimum air temperature, 
        :math:`e_{s,min}`
        [mbar]

    """
    return saturated_vapour_pressure(t_air_min_coarse)


def saturated_vapour_pressure_maximum(t_air_max_coarse):
    """Like :func:`saturated_vapour_pressure` but based on daily maximum air temperature. This
    is only relevant for reference ET calculations.

    Parameters
    ----------
    t_air_max_coarse : float
        daily maximum air temperature, 
        :math:`T_{a,max}`
        [C]

    Returns
    -------
    svp_24_max : float
        daily saturated vapour pressure based on maximum air temperature, 
        :math:`e_{s,max}`
        [mbar]

    """
    return saturated_vapour_pressure(t_air_max_coarse)


def saturated_vapour_pressure_average(svp_24_max, svp_24_min):
    r"""
    Average saturated vapour pressure based on two saturated vapour pressure values
    calculated using minimum and maximum air temperature respectively. This is preferable
    to calculating saturated vapour pressure using the average air temperature, because
    of the strong non-linear relationship between saturated vapour pressure and air
    temperature.

    .. math ::

        e_{s}=\frac{e^{0}\left(T_{max}\right)+e^{0}\left(T_{min}\right)}{2}

    Parameters
    ----------
    svp_24_max : float
        daily saturated vapour pressure based on maximum air temperature, 
        :math:`e_{s,max}`
        [mbar]
    svp_24_min : float
        daily saturated vapour pressure based on minimum air temperature, 
        :math:`e_{s,min}`
        [mbar]

    Returns
    -------
    svp_24 : float
        daily saturated vapour pressure, 
        :math:`e_{s,24}`
        [mbar]

    """
    return (svp_24_max + svp_24_min)/2


def vapour_pressure_from_specific_humidity_inst(qv_i, p_air_i):
    """Like :func:`vapour_pressure_from_specific_humidity` but as an instantaneous value.

    Parameters
    ----------
    qv_i : float
        instantaneous specific humidity, 
        :math:`q_{v,i}`
        [kg/kg]
    p_air_i : float
        instantaneous air pressure, 
        :math:`P_{i}`
        [mbar]

    Returns
    -------
    vp_i : float
        instantaneous vapour pressure, 
        :math:`e_{a,i}`
        [mbar]
    """
    return vapour_pressure_from_specific_humidity(qv_i, p_air_i)


def saturated_vapour_pressure(t_air):
    r"""
    Computes saturated vapour pressure :math:`e_s` [mbar], it provides
    the vapour pressure when the air is fully saturated with water. It is
    related to air temperature :math:`T_a` [C] as:

    .. math ::

        e_{s}=6.108 \cdot \exp\left[\frac{17.27 \cdot T_{a}}{T_{a}+237.3}\right]

    Parameters
    ----------
    t_air : float
        air temperature, 
        :math:`T_a`
        [C]

    Returns
    -------
    svp : float
        saturated vapour pressure, 
        :math:`e_s`
        [mbar]

    Examples
    --------
    >>> from ETLook import meteo
    >>> meteo.saturated_vapour_pressure(20)
    23.382812709274457

    """
    return 6.108 * np.exp(((17.27 * t_air) / (237.3 + t_air)))


def saturated_vapour_pressure_daily(t_air_24):
    """Like :func:`saturated_vapour_pressure` but as a daily average.

    Parameters
    ----------
    t_air_24 : float
        daily air temperature, 
        :math:`T_{a,24}`
        [C]

    Returns
    -------
    svp_24 : float
        daily saturated vapour pressure, 
        :math:`e_{s,24}`
        [mbar]

    """
    return saturated_vapour_pressure(t_air_24)


def saturated_vapour_pressure_inst(t_air_i):
    """Like :func:`saturated_vapour_pressure` but as an instantaneous value.

    Parameters
    ----------
    t_air_i : float
        instantaneous air temperature, 
        :math:`T_{a,i}`
        [C]

    Returns
    -------
    svp_i : float
        instantaneous saturated vapour pressure, 
        :math:`e_{s,i}`
        [mbar]

    """
    return saturated_vapour_pressure(t_air_i)


def vapour_pressure(svp, rh):
    r"""
    Computes the vapour pressure :math:`e_a` in [mbar].

    .. math ::

        e_{a}=\frac{\phi}{100} \cdot e_{s}

    Parameters
    ----------
    svp : float
        saturated vapour pressure, 
        :math:`e_s`
        [mbar]
    rh : float
        relative humidity, 
        :math:`\phi`
        [%]

    Returns
    -------
    vp : float
        vapour pressure, 
        :math:`e_{a}`
        [mbar]

    Examples
    --------
    >>> from ETLook import meteo
    >>> meteo.vapour_pressure(rh=75, svp=meteo.saturated_vapour_pressure(20))
    17.537109531955842

    """
    return 0.01 * rh * svp


def slope_saturated_vapour_pressure(t_air):
    r"""
    Computes the rate of change of vapour pressure :math:`\Delta` in [mbar K-1]
    for a given air temperature :math:`T_a`. It is a function of the air
    temperature :math:`T_a` and the saturated vapour pressure :math:`e_s`
    [mbar] which in itself is a function of :math:`T_a`.

    .. math ::

        \Delta=\frac{4098 \cdot e_{s}}{\left(237.3+T_{a}\right)^{2}}

    for :math:`e_s` see :func:`saturated_vapour_pressure`

    Parameters
    ----------
    t_air : float
       air temperature
       :math:`T_a`
       [C]

    Returns
    -------
    ssvp : float
       slope of saturated vapour pressure curve
       :math:`\Delta`
       [mbar K-1]

    Examples
    --------
    >>> from ETLook import meteo
    >>> meteo.slope_saturated_vapour_pressure(20)
    1.447401881124136

    """
    svp = saturated_vapour_pressure(t_air)
    return (4098 * svp) / (237.3 + t_air) ** 2


def slope_saturated_vapour_pressure_daily(t_air_24):
    """Like :func:`slope_saturated_vapour_pressure` but as a daily average.

    Parameters
    ----------
    t_air_24 : float
       daily air temperature
       :math:`T_{a,24}`
       [C]

    Returns
    -------
    ssvp_24 : float
       daily slope of saturated vapour pressure curve
       :math:`\Delta_{24}`
       [mbar K-1]


    """
    return slope_saturated_vapour_pressure(t_air_24)


def slope_saturated_vapour_pressure_inst(t_air_i):
    """Like :func:`slope_saturated_vapour_pressure` but as an instantaneous.
    value

    Parameters
    ----------
    t_air_i : float
        instantaneous air temperature, 
        :math:`T_{a,i}`
        [C]

    Returns
    -------
    ssvp_i : float
        instantaneous slope of saturated vapour pressure curve, 
        :math:`e_{s,i}`
        [mbar]
    """
    return slope_saturated_vapour_pressure(t_air_i)


def vapour_pressure_deficit(svp, vp):
    r"""
    Computes the vapour pressure deficit :math:`\Delta_e` in [mbar].

    .. math ::

        \Delta_e=e_s-e_a

    Parameters
    ----------
    svp : float
        saturated vapour pressure, 
        :math:`e_s`
        [mbar]
    vp : float
        actual vapour pressure, 
        :math:`e_a`
        [mbar]

    Returns
    -------
    vpd : float
       vapour pressure deficit
       :math:`\Delta_e`
       [mbar]

    Examples
    --------
    >>> from ETLook import meteo
    >>> meteo.vapour_pressure_deficit(12.5, 5.4)
    7.1
    >>> meteo.vapour_pressure_deficit(vp=5.4, svp=12.3)
    6.9

    """
    vpd = svp - vp
    if isinstance(vpd, xr.DataArray):
        vpd = xr.where(vpd < 0, 0.0, vpd)
    else:
        vpd[vpd < 0] = 0

    return vpd


def vapour_pressure_deficit_daily(svp_24, vp_24):
    """Like :func:`vapour_pressure_deficit` but as a daily average.

    Parameters
    ----------
    svp_24 : float
        daily saturated vapour pressure, 
        :math:`e_{s,24}`
        [mbar]
    vp_24 : float
        daily actual vapour pressure, 
        :math:`e_{a,24}`
        [mbar]

    Returns
    -------
    vpd_24 : float
       daily vapour pressure deficit
       :math:`\Delta_{e,24}`
       [mbar]
    """
    return vapour_pressure_deficit(svp_24, vp_24)


def air_pressure(z, p_air_0=1013.25):
    r"""
    Computes air pressure :math:`P` at a certain elevation derived from the
    air pressure at sea level :math:`P_0`. Air pressure decreases with
    increasing elevation.

    .. math ::

        P=P_{0} \cdot \left(\frac{T_{ref,0,K}-\alpha_{1} \cdot \left(z-z_{0}\right)}
        {T_{ref,0,K}}\right)^{\frac{g}{-\alpha_{1}\cdot R }}

    where the following constants are used

    * :math:`P_0` = air pressure [mbar] at sea level :math:`z_0` = 1013.25 mbar
    * :math:`T_{ref,0,K}` = reference temperature [K] at sea level
      :math:`z_0` = 293.15 K
    * :math:`g` = gravitational acceleration = 9.807 [m/s2]
    * :math:`R` = specific gas constant = 287.0 [J kg-1 K-1]
    * :math:`\alpha_{1}` = constant lapse rate for moist air = 0.0065 [K m-1]

    Parameters
    ----------
    z : float
        elevation, 
        :math:`z`
        [m]
    p_air_0 : float
        air pressure at sea level, 
        :math:`P_0`
        [mbar]

    Returns
    -------
    p_air : float
        air pressure, 
        :math:`P`
        [mbar]

    Examples
    --------
    >>> from ETLook import meteo
    >>> meteo.air_pressure(z=1000)
    900.5832172948869

    """
    return p_air_0 * ((c.t_ref + c.lapse * (z - c.z_ref)) / c.t_ref) ** c.power


def air_pressure_daily(z, p_air_0_24=1013.25):
    r"""Like :func:`air_pressure` but as a daily average.

    Parameters
    ----------
    z : float
        elevation, 
        :math:`z`
        [m]
    p_air_0_24 : float
        daily air pressure at sea level, 
        :math:`P_{0,24}`
        [mbar]

    Returns
    -------
    p_air_24 : float
        daily air pressure, 
        :math:`P_{24}`
        [mbar]

    """
    return air_pressure(z, p_air_0_24)


def dry_air_density(p_air, vp, t_air_k):
    r"""
    Computes dry air density :math:`\rho_{d}` in [kg m-3].

    .. math ::

        \rho_{d}=\frac{P-e_{a}}{\Re \cdot T_{a,K}}

    where the following constants are used

    * :math:`\Re` = gas constant for dry air = 2.87 mbar K-1 m3 kg-1

    Parameters
    ----------
    p_air : float
        air pressure, 
        :math:`P`
        [mbar]
    vp : float
        vapour pressure, 
        :math:`e_{a}`
        [mbar]
    t_air_k : float
        daily air temperature, 
        :math:`T_{a}`
        [K]

    Returns
    -------
    ad_dry : float
        dry air density, 
        :math:`\rho_{d}`
        [kg m-3]

    Examples
    --------
    >>> from ETLook import meteo
    >>> meteo.dry_air_density(p_air=900, vp=17.5, t_air_k=293.15)
    1.0489213344656534

    """
    return (p_air - vp) / (t_air_k * c.gc_dry)


def dry_air_density_daily(p_air_24, vp_24, t_air_k_24):
    r"""
    Like :func:`dry_air_density` but as a daily average.

    Parameters
    ----------
    p_air_24 : float
        daily air pressure, 
        :math:`P_{24}`
        [mbar]
    vp_24 : float
        daily vapour pressure, 
        :math:`e_{a,24}`
        [mbar]
    t_air_k_24 : float
        daily air temperature, 
        :math:`T_{a,24}`
        [K]

    Returns
    -------
    ad_dry_24 : float
        daily dry air density, 
        :math:`\rho_{d,24}`
        [kg m-3]
    """
    return dry_air_density(p_air_24, vp_24, t_air_k_24)


def dry_air_density_inst(p_air_i, vp_i, t_air_k_i):
    r"""
    Like :func:`dry_air_density` but as an instantaneous value.

    Parameters
    ----------
    p_air_i : float
        instantaneous air pressure, 
        :math:`P_{i}`
        [mbar]
    vp_i : float
        instantaneous vapour pressure, 
        :math:`e_{a,i}`
        [mbar]
    t_air_k_i : float
        instantaneous air temperature, 
        :math:`T_{a,i}`
        [K]

    Returns
    -------
    ad_dry_i : float
        instantaneous dry air density, 
        :math:`\rho_{d,i}`
        [kg m-3]
    """
    return dry_air_density(p_air_i, vp_i, t_air_k_i)


def moist_air_density(vp, t_air_k):
    r"""
    Computes moist air density :math:`\rho_{s}` in [kg m-3].

    .. math ::

        \rho_{s}=\frac{e_{a}}{R_{v} \cdot T_{a,K}}

    where:

    * :math:`R_v` = gas constant for moist air = 4.61 mbar K-1 m3 kg-1

    Parameters
    ----------
    vp : float
        vapour pressure, 
        :math:`e_{a}`
        [mbar]
    t_air_k : float
        air temperature, 
        :math:`T_{a,K}`
        [K]

    Returns
    -------
    ad_moist : float
        moist air density, 
        :math:`\rho_{s}`
        [kg m-3]

    Examples
    --------
    >>> from ETLook import meteo
    >>> meteo.moist_air_density(vp=17.5, t_air_k = 293.15)
    0.012949327800393881

    """
    return vp / (t_air_k * c.gc_moist)


def moist_air_density_daily(vp_24, t_air_k_24):
    r"""
    Like :func:`moist_air_density` but as a daily average.

    Parameters
    ----------
    vp_24 : float
        daily vapour pressure, 
        :math:`e_{a,24}`
        [mbar]
    t_air_k_24 : float
        daily air temperature, 
        :math:`T_{a,K,24}`
        [K]

    Returns
    -------
    ad_moist_24 : float
        daily moist air density, 
        :math:`\rho_{s,24}`
        [kg m-3]

    """
    return moist_air_density(vp_24, t_air_k_24)


def moist_air_density_inst(vp_i, t_air_k_i):
    r"""
    Like :func:`moist_air_density` but as an instantaneous value.

    Parameters
    ----------
    vp_i : float
        instantaneous vapour pressure, 
        :math:`e_{a,i}`
        [mbar]
    t_air_k_i : float
        instantaneous air temperature, 
        :math:`T_{a,K,i}`
        [K]

    Returns
    -------
    ad_moist_i : float
        instantaneous moist air density, 
        :math:`\rho_{s,i}`
        [kg m-3]

    """
    return moist_air_density(vp_i, t_air_k_i)


def air_density(ad_dry, ad_moist):
    r"""
    Computes air density :math:`\rho` in [kg m-3].

    .. math ::

        \rho=\rho_{s}+\rho_{d}

    Parameters
    ----------
    ad_dry : float
        dry air density, 
        :math:`\rho_{d}`
        [kg m-3]
    ad_moist : float
        moist air density, 
        :math:`\rho_{s}`
        [kg m-3]

    Returns
    -------
    ad : float
        air density, 
        :math:`\rho`
        [kg m-3]

    Examples
    --------
    >>> from ETLook import meteo
    >>> ad_moist = meteo.moist_air_density(vp=17.5, t_air_k = 293.15)
    >>> ad_dry = meteo.dry_air_density(p_air=900, vp=17.5, t_air_k=293.15)
    >>> meteo.air_density(ad_dry=ad_dry, ad_moist=ad_moist)
    1.0618706622660472

    """
    return ad_dry + ad_moist


def air_density_daily(ad_dry_24, ad_moist_24):
    r"""

    Like :func:`air_density` but as a daily average.

    Parameters
    ----------
    ad_dry_24 : float
        daily dry air density, 
        :math:`\rho_{d,24}`
        [kg m-3]
    ad_moist_24 : float
        daily moist air density, 
        :math:`\rho_{s,24}`
        [kg m-3]

    Returns
    -------
    ad_24 : float
        daily air density, 
        :math:`\rho_{24}`
        [kg m-3]

    """
    return air_density(ad_dry_24, ad_moist_24)


def air_density_inst(ad_dry_i, ad_moist_i):
    r"""
    Like :func:`air_density` but as a instantaneous value.

    Parameters
    ----------
    ad_dry_i : float
        instantaneous dry air density, 
        :math:`\rho_{d,i}`
        [kg m-3]
    ad_moist_i : float
        instantaneous moist air density, 
        :math:`\rho_{s,i}`
        [kg m-3]

    Returns
    -------
    ad_i : float
        instantaneous air density, 
        :math:`\rho_{i}`
        [kg m-3]

    """
    return air_density(ad_dry_i, ad_moist_i)


def latent_heat(t_air):
    r"""
    Computes latent heat of evaporation :math:`\lambda` [J kg-1], describing
    the amount of energy needed to evaporate one kg of water at constant
    pressure and temperature. At higher temperatures less energy will be
    required than at lower temperatures.

    .. math ::

        \lambda=\lambda_0 + \Delta_\lambda \cdot T_{a}

    where the following constants are used

    * :math:`\lambda_0` = latent heat of evaporation at 0 C = 2501000 [J kg-1]
    * :math:`\Delta_\lambda` = rate of change of latent heat with respect to temperature = -2361 [J Kg-1 C-1]

    Parameters
    ----------
    t_air : float
        air temperature, 
        :math:`T_a`
        [C]

    Returns
    -------
    lh : float
        latent heat of evaporation, 
        :math:`\lambda`
        [J/kg]

    Examples
    --------
    >>> from ETLook import meteo
    >>> meteo.latent_heat(20)
    2453780.0

    """
    return c.lh_0 + c.lh_rate * t_air


def latent_heat_daily(t_air_24):
    """Like :func:`latent_heat` but as a daily average.

    Parameters
    ----------
    t_air_24 : float
        daily air temperature, 
        :math:`T_{a,24}`
        [C]

    Returns
    -------
    lh_24 : float
        daily latent heat of evaporation, 
        :math:`\lambda_{24}`
        [J/kg]

    """
    return latent_heat(t_air_24)


def psychrometric_constant(p_air, lh):
    r"""
    Computes the psychrometric constant :math:`\gamma` [mbar K-1] which
    relates the partial pressure of water in air to the air temperature.

    .. math ::

        \gamma=\frac{P \cdot c_{p}}{\varepsilon \cdot \lambda}

    where the following constants are used

    * :math:`c_{p}` = specific heat for dry air = 1004 [J Kg-1 K-1]
    * :math:`\varepsilon` = ratio of molecular weight of water to
      dry air = 0.622 [-]

    Parameters
    ----------
    p_air : float
        air pressure, 
        :math:`P`
        [mbar]
    lh : float
        latent heat of evaporation, 
        :math:`\lambda`
        [J/kg]

    Returns
    -------
    psy : float
        psychrometric constant, 
        :math:`\gamma`
        [mbar K-1]

    Examples
    --------
    >>> from ETLook import meteo
    >>> meteo.psychrometric_constant(p_air = 1003.0, lh = 2500000.0)
    0.6475961414790997
    >>> meteo.psychrometric_constant(1003.0, 2500000.0)
    0.6475961414790997

    """
    return (c.sh * p_air) / (c.r_mw * lh)


def psychrometric_constant_daily(p_air_24, lh_24):
    """Like :func:`psychrometric_constant` but as a daily average.

    Parameters
    ----------
    p_air_24 : float
        daily air pressure, 
        :math:`P_{24}`
        [mbar]
    lh_24 : float
        daily latent heat of evaporation, 
        :math:`\lambda_{24}`
        [J/kg]

    Returns
    -------
    psy_24 : float
        daily psychrometric constant, 
        :math:`\gamma_{24}`
        [mbar K-1]

    """
    return psychrometric_constant(p_air_24, lh_24)


def wind_speed_blending_height(u, z_obs=2, z_b=100):
    r"""
    Computes the wind speed at blending height :math:`u_{b}` [m/s] using the
    logarithmic wind profile.

    .. math ::

        u_{b}=\frac{u_{obs} \cdot \ln\left(\frac{z_{b}}{z_{0,m}}\right)}
        {\ln\left(\frac{z_{obs}}{z_{0,m}}\right)}

    Parameters
    ----------
    u : float
        wind speed at observation height, 
        :math:`u_{obs}`
        [m/s]
    z_obs : float
        observation height of wind speed, 
        :math:`z_{obs}`
        [m]
    z_b : float
        blending height, 
        :math:`z_{b}`
        [m]

    Returns
    -------
    u_b : float
        wind speed at blending height, 
        :math:`u_{b}`
        [m/s]

    Examples
    --------
    >>> from ETLook import meteo
    >>> meteo.wind_speed_blending_height(u=3.0, z_obs=2, z_b=100)
    5.4646162953650572

    """

    z0m = 0.0171

    ws = (c.k * u) / np.log(z_obs / z0m) * np.log(z_b / z0m) / c.k

    if isinstance(ws, xr.DataArray):
        ws = ws.clip(1, 150)
    else:
        ws = np.clip(ws, 1, 150)    

    return ws


def wind_speed_blending_height_daily(u_24, z_obs=2, z_b=100):
    """Like :func:`wind_speed_blending_height` but as a daily average.

    Parameters
    ----------
    u_24 : float
        daily wind speed at observation height, 
        :math:`u_{obs,24}`
        [m/s]
    z_obs : float
        observation height of wind speed, 
        :math:`z_{obs}`
        [m]
    z_b : float
        blending height, 
        :math:`z_{b}`
        [m]

    Returns
    -------
    u_b_24 : float
        daily wind speed at blending height, 
        :math:`u_{b, 24}`
        [m/s]

    """
    return wind_speed_blending_height(u_24, z_obs, z_b)


def wind_speed(u, v):
    """Calculate wind speed vector from two components.

    Parameters
    ----------
    u : float
        Eastward wind speed.
    v : float
        Northward wind speed.

    Returns
    -------
    u_24
        Wind speed.
    """
    return np.sqrt(v**2 + u**2)


def air_pressure_kpa2mbar(p_air_kpa):
    """Like :func:`p_air`.

    Parameters
    ----------
    p_air_kpa : float
        air pressure, 
        :math:`Pair_{a}`
        [kpa]

    Returns
    -------
    p_air_mbar : float
        air pressure, 
        :math:`Pair_{a}`
        [mbar]
    """
    return p_air_kpa * 10